# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

from __future__ import print_function
from ..utils import clingo_stats
from ..utils import utils
from ..utils import clingo_signal_handler
import sys

BASE = utils.BASE
WARNING_INCLUDED_FILE = "<cmd>: warning: already included file:\n  {}\n"
ERROR_INCLUDED_FILE = "file could not be opened:\n  {}\n"
MESSAGE_LIMIT = 20
TOO_MANY = "too many messages."
INTERRUPT = clingo_signal_handler.INTERRUPT.format("asprin")
SUMMARY_STR = clingo_signal_handler.SUMMARY_STR
STATS_STR = clingo_signal_handler.STATS_STR

class Printer:

    messages = 0  # class variables
    last     = "" #

    #
    # errors and warnings
    #
   
    def __check_messages(self, n):
        Printer.messages += n
        if Printer.messages >= MESSAGE_LIMIT:
            pass #raise Exception(TOO_MANY) 

    def __last(self, string):
        if Printer.last == string:
            return True
        Printer.last = string
        return False

    def __print_error(self, string, **kwargs):
        if not self.__last(string):
            sys.stdout.flush()
            print(string, file=sys.stderr, **kwargs)
            self.__check_messages(1)

    def print_error_string(self, string):
        if not self.__last(string):
            sys.stdout.flush()
            print(string, file=sys.stderr, end = "")
            self.__check_messages(int(string.count("\n")//2))

    def print_error_location(self, location, string):
        self.__print_error("{}error: {}".format(location,string))

    def print_error(self, string):
        self.__print_error(string)

    def print_spec_error(self, string):
        self.__print_error(string)

    def print_warning(self, string, **kwargs):
        if not self.__last(string):
            sys.stdout.flush()
            print(string, file=sys.stderr, **kwargs)
            self.__check_messages(1)

    def print_spec_warning(self, string):
        self.print_warning(string)

    def warning_included_file(self, file, loc=None):
        warning = WARNING_INCLUDED_FILE 
        if loc: 
            warning = warning.replace("<cmd>: ", str(loc))
        self.print_warning(warning.format(file))

    def error_included_file(self, file, loc):
        self.print_error_location(loc, ERROR_INCLUDED_FILE.format(file))

    #
    # simply print
    #
    def do_print(self, *args, **kwargs):
        print(*args, **kwargs)

    #
    # stats
    #

    def print_stats(self, ctl, models, more_models,
                    opt_models, non_optimal, stats,
                    interrupted, solved, copy_statistics):
        # interrupt
        out = ""
        if interrupted:
            out = INTERRUPT
        # first lines
        out += "\nModels       : {}{}".format(models,"+" if more_models else "")
        if not non_optimal:
            out += "\n  Optimum    : {}".format("yes" if opt_models>0 else "no")
            if opt_models > 0:
                out += "\n  Optimal    : {}".format(opt_models)
        out += "\n"
        # gather string
        if not solved:
            out += SUMMARY_STR + STATS_STR if stats else SUMMARY_STR
        else:
            # copy statistics dictionary
            statistics = ctl.statistics
            if copy_statistics is not None:
                statistics = copy_statistics
            # add statistics
            out += clingo_stats.Stats().summary(statistics, False)
            if stats:
                out += "\n" + clingo_stats.Stats().statistics(statistics)
        # print
        print(out)
        sys.stdout.flush()




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

from __future__ import print_function
from collections import namedtuple


NodeInfo = namedtuple('NodeInfo','key item')


class Node:

    def __init__(self, key, item):
        self.key  = key
        self.item = item
        self.next = set()
        self.prev = set()
        self.neg_next = set()
        self.neg_prev = set()

    def __str__(self):
        out = []
        if self.next:
            out += [(i.key, "+") for i in self.next]
        if self.neg_next:
            out += [(i.key, "-") for i in self.neg_next]
        ret = "#{}\n:{}\n".format(self.key, str(self.item))
        list_ = ["({},{},{})".format(self.key, i[0], i[1]) for i in out]
        return ret + "\n".join(list_)


class TransitiveClosure:

    def __init__(self):
        self.nodes = {}

    #
    # CREATE THE GRAPH
    #

    # add set_1 to set_2, and delete set_1 from set_3
    def __update(self, set_1, set_2, set_3):
        set_2.update(set_1)
        set_3.difference_update(set_1)

    # update graph with (NodeInfo) a
    # do not add item if it is None
    def add_node(self, a):
        node = self.nodes.get(a.key)
        if not node:
            item = [a.item] if a.item is not None else []
            node = Node(a.key, item)
            self.nodes[a.key] = node
        elif a.item is not None:
            node.item.append(a.item)
        return node

    # add edge from (NodeInfo) a to (NodeInfo) b 
    # if flag, then the edge has negative sign
    # if not add_node, then a and b must be in the graph
    def add_edge(self, a, b, flag, add_node=False):

        # nodes
        if add_node:
            node_a = self.add_node(a)
            node_b = self.add_node(b)
        else:
            node_a = self.nodes[a.key]
            node_b = self.nodes[b.key]

        # next
        if not flag: # positive sign
            next = node_b.next.copy()
            next.add(node_b)
            node_a.next.update(next)
            for i in node_a.prev:
                i.next.update(next)
            for i in node_a.neg_prev:
                self.__update(next, i.neg_next, i.next)

        # neg_next
        if not flag: # positive sign
            neg_next = node_b.neg_next
        else:
            neg_next = node_b.neg_next.union(node_b.next)
            neg_next.add(node_b)
        if neg_next:
            self.__update(neg_next, node_a.neg_next, node_a.next)
            for i in node_a.prev:
                self.__update(neg_next, i.neg_next, i.next)
            for i in node_a.neg_prev:
                self.__update(neg_next, i.neg_next, i.next)

        # prev
        if not flag: # positive sign
            prev = node_a.prev.copy()
            prev.add(node_a)
            node_b.prev.update(prev)
            for i in node_b.next:
                i.prev.update(prev)
            for i in node_b.neg_next:
                self.__update(prev, i.neg_prev, i.prev)

        # neg_prev
        if not flag: # positive sign
            neg_prev = node_a.neg_prev
        else:
            neg_prev = node_a.neg_prev.union(node_a.prev)
            neg_prev.add(node_a)
        if neg_prev:
            self.__update(neg_prev, node_b.neg_prev, node_b.prev)
            for i in node_b.next:
                self.__update(neg_prev, i.neg_prev, i.prev)
            for i in node_a.neg_next:
                self.__update(neg_prev, i.neg_prev, i.prev)

    def __str__(self):
        out = ""
        for key, item in self.nodes.items():
            out += str(item) + "\n"
        return out

    #
    # USE THE GRAPH
    #

    # pre: key must be in the graph
    def get_next(self, key):
        out = self.nodes[key].next.union(self.nodes[key].neg_next)
        return [i.key for i in out]

    def get_cycles(self):
        out = []
        for key, value in self.nodes.items():
            if value in value.neg_next:
                out.append(NodeInfo(key, None))
        return out

    def map_items(self, f):
        for node in self.nodes.values():
            for i in node.item:
                f(i)


if __name__ == "__main__":
    graph = [(1,2,True), (2,3,True), (3,4,False), (4,5,True), (5,5,True),
             (7,8,False), (8,7,False), (2,1,False)]#, (5,1,True)]
    tmp = []
    for i in range(1,2):
        for j in graph:
            tmp.append((j[0]*i,j[1]*i,j[2]))
    graph = tmp
    tc = TransitiveClosure()
    for i in graph:
        tc.add_edge(NodeInfo(i[0],i[0]), NodeInfo(i[1],i[1]), i[2], True)
    print(tc)




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

from ..utils import utils

class GeneralController:

    def __init__(self, solver):
        self.improve_limit = True if solver.options.improve_limit is not None \
            else False
        self.solver = solver
        self.bool_set_unsat_program = True

    def start(self):
        solver, options = self.solver, self.solver.options
        if options.trans_ext is not None:
            solver.control.configuration.asp.trans_ext = options.trans_ext
        # store nholds and set holds domain
        if solver.options.max_models != 1:
            solver.store_nholds = True
        if solver.store_nholds:
            solver.set_holds_domain = True
        if solver.set_holds_domain:
            solver.do_set_holds_domain()
        #
        solver.add_encodings()
        solver.ground_preference_base()
        # --dom-heur
        if options.cmd_heuristic is not None:
            solver.ground_cmd_heuristic()
            for _solver in solver.control.configuration.solver:
                _solver.heuristic="Domain"
        # check syntax
        if options.check:
            solver.check_errors()
        # non optimal
        if solver.no_optimize() or solver.options.non_optimal:
            solver.str_found      = utils.STR_MODEL_FOUND
            solver.str_found_star = utils.STR_MODEL_FOUND_STAR
            # solve and finish asprin
            self.solver.solve_single()
            self.solver.end()

    def sat(self):
        self.solver.models     += 1
        self.solver.last_unsat  = False
        self.solver.last_model  = self.solver.step
        self.solver.check_last_model()
        if self.solver.options.quiet == 0:
            self.solver.print_answer()
        else:
            self.solver.print_str_answer()

    def unsat(self):
        if self.solver.last_unsat:
            self.solver.more_models = False
            if self.solver.models == 0:
                self.solver.print_unsat()
            elif self.improve_limit:
                self.solver.enumerate_unknown()
            self.solver.end()
        if self.solver.options.quiet == 1:
            self.solver.print_shown()
        self.solver.last_unsat = True
        self.solver.opt_models  += 1
        self.solver.print_optimum_string()
        if self.solver.opt_models == self.solver.options.max_models:
            self.solver.end()
        if self.solver.options.steps == self.solver.step:
            self.solver.print_steps_message()
            self.solver.end() # to exit asap in this case
        # set_unsat_program
        if self.bool_set_unsat_program:
            self.solver.set_unsat_program()
            self.bool_set_unsat_program = False

    def unknown(self):
        if self.solver.last_unsat:
            raise utils.FatalException
        self.solver.print_limit_string()
        self.solver.last_unsat = True # act as if unsat
        if self.solver.options.steps == self.solver.step:
            self.solver.print_steps_message()
            self.solver.end() # to exit asap in this case
        # set_unsat_program
        if self.bool_set_unsat_program:
            self.solver.set_unsat_program()
            self.bool_set_unsat_program = False

    def end_loop(self):
        self.solver.step = self.solver.step + 1
        if self.solver.options.steps == (self.solver.step - 1):
            self.solver.print_steps_message()
            self.solver.end()
        if self.solver.options.clean_up:
            self.solver.clean_up()


class GeneralControllerHandleOptimal:

    def __init__(self, solver):
        self.solver = solver
        self.start_step = True
        self.first  = True
        self.delete_worse  = True
        self.delete_better = self.solver.options.delete_better
        self.total_order   = self.solver.options.total_order
        self.no_opt_improving, self.volatile = False, False
        if solver.options.no_opt_improving:
            if solver.options.max_models != 1:
                self.no_opt_improving = True
            self.volatile = True
        if solver.options.volatile_optimal:
            self.volatile = True

    def start(self):
        if self.delete_better:
            self.solver.ground_holds_delete_better()

    def sat(self):
        if self.no_opt_improving and self.start_step:
            self.solver.relax_optimal_models()
        self.start_step = False

    def __preprocess(self):
        if self.total_order and not self.first:
            self.delete_worse  = False
            self.delete_better = False
        self.first = False

    def unsat(self):
        self.start_step = True
        self.__preprocess()
        self.solver.handle_optimal_model(self.solver.last_model,
                                         False,
                                         self.delete_worse,
                                         self.delete_better,
                                         self.volatile)

    def unknown(self):
        self.start_step = True
        self.__preprocess()
        self.solver.handle_optimal_model(self.solver.last_model,
                                         True,
                                         self.delete_worse,
                                         False,
                                         True)


class EnumerationController:

    def __init__(self, solver):
        self.solver = solver
        # show_underscores is set by program_parser.py
        if not solver.options.project and not solver.options.show_underscores:
            solver.same_shown_function = solver.same_shown

    def unsat(self):
        if not self.solver.options.project:
            self.solver.enumerate()
            if self.solver.opt_models == self.solver.options.max_models:
                self.solver.end()


#
# Method Controllers
#

class MethodController:

    def __init__(self, solver):
        self.solver = solver

    def start(self):
        pass

    def start_loop(self):
        pass

    def solve(self):
        pass

    def unsat(self):
        pass


class GroundManyMethodController(MethodController):

    def __init__(self, solver):
        MethodController.__init__(self, solver)
        self.volatile = False
        self.ground_holds = 0
        if (self.solver.options.max_models != 1 or
            self.solver.options.release_last or 
            self.solver.options.volatile_improving):
            self.volatile = True 

    def start_loop(self):
        if not self.solver.last_unsat:
            self.solver.ground_holds(self.solver.last_model)
            self.ground_holds = self.solver.last_model
            self.solver.ground_preference_program(self.volatile)

    def solve(self):
        self.solver.solve(
            assumptions=self.solver.assumptions,
            on_model=self.solver.on_model
        )

    def unsat(self):
        self.solver.relax_previous_models()
        if self.ground_holds != self.solver.last_model:
            self.solver.ground_holds(self.solver.last_model)


class GroundOnceMethodController(MethodController):

    def __init__(self, solver):
        MethodController.__init__(self, solver)
        solver.store_nholds = True

    def start(self):
        self.solver.ground_open_preference_program()
        self.solver.turn_off_preference_program()

    def start_loop(self):
        if not self.solver.last_unsat:
            self.solver.turn_on_preference_program()

    def solve(self):
        self.solver.solve(
            assumptions=self.solver.assumptions,
            on_model=self.solver.on_model
        )

    def unsat(self):
        self.solver.ground_holds(self.solver.last_model)
        self.solver.turn_off_preference_program()


class ApproxMethodController(MethodController):

    def __init__(self, solver):
        MethodController.__init__(self, solver)
        self.solver.store_nholds = True

    def start(self):
        self.solver.solve_approx()
        # finishes asprin


class HeurMethodController(MethodController):

    def __init__(self, solver):
        MethodController.__init__(self, solver)

    def start(self):
        self.solver.ground_heuristic()

    def solve(self):
        if self.solver.last_unsat:
            self.solver.solve_heuristic()
        else:
            self.solver.solve_unsat()

    def unsat(self):
        self.solver.ground_holds(self.solver.last_model)


class MetaMethodController(MethodController):

    def __init__(self, solver):
        MethodController.__init__(self, solver)
        self.solver = solver
        self.solver.set_holds_domain = True

    def start(self):
        # run meta
        self.solver.meta_simple()
        # finishes asprin
        self.solver.end()


class ImproveLimitController(MethodController):

    def __init__(self, solver, controller):
        MethodController.__init__(self, solver)
        self.solver     = solver
        self.controller = controller
        self.option     = self.solver.options.improve_limit
        self.stats      = solver.control.statistics['solving']['solvers']
        self.conf       = solver.control.configuration.solve
        self.search     = 0
        # limits
        self.previous_limit, self.limit = "", 0
        # store nholds
        self.solver.store_nholds = True

    def start(self):
        self.controller.start()
        self.solver.ground_holds_delete_better()

    def start_loop(self):
        # get previous limit, and set limit
        self.previous_limit = self.conf.solve_limit
        if not self.solver.last_unsat:
            self.limit = (1 if self.search==0 else self.search) * self.option[0]
            if self.limit < self.option[2]:
                self.limit = self.option[2]
            self.conf.solve_limit = str(self.limit) + ",umax"
            if self.limit == 0: # if limit is 0, return
                return
        # call controller
        self.controller.start_loop()

    def solve(self):
        # if improving and limit is 0, return unknown
        if not self.solver.last_unsat and self.limit == 0:
            self.solver.solve_unknown()
        else:
            self.controller.solve()
        # reset previous limit
        self.conf.solve_limit = self.previous_limit
        # gather search results
        if self.solver.solving_result == utils.SATISFIABLE:
            if self.solver.last_unsat:
                self.search  = int(self.stats['conflicts'])
            elif self.option[1]:
                self.search += int(self.stats['conflicts'])

    def unsat(self):
        result = self.solver.solving_result
        self.controller.unsat()
        self.solver.handle_unknown_models(result)


#
# class OnOptimal
#

# defines (from utils)
HOLDS      = utils.HOLDS
LAST_HOLDS = utils.LAST_HOLDS
LAST_SHOWN = utils.LAST_SHOWN
POS        = utils.POS
NEG        = utils.NEG
SHOWN_ATOM = utils.SHOWN_ATOM
PREF_ATOM  = utils.PREF_ATOM
ON_OPT_HEUR_PROGRAM = utils.ON_OPT_HEUR_PROGRAM

# heuristic and external rules
HEUR_RULE = """
#heuristic ##""" + HOLDS + """(X,0) : not ##""" + LAST_HOLDS + """(X). [#v,#m]
"""
HEUR_EXTERNAL_HOLDS = """
#external ##""" + LAST_HOLDS + """(X) : X = @get_holds_domain().
"""
HEUR_EXTERNAL_SHOWN = """
#external ##""" + LAST_SHOWN + """(X) : X = @get_shown_domain().
"""

#
# Note: Option --on-opt-heur does not change anything while enumerating
#       optimal models with the same preference atoms (OPTIMUM FOUND *)
#

class OnOptimalController:

    def __init__(self, solver):
        self.solver     = solver
        self.grounded   = False
        self.heuristic  = False
        self.pref       = False
        self.shown      = False
        self.shown_rule = ""
        # for on_opt_heur
        if self.solver.options.on_opt_heur:
            self.heuristic = True
            # activate domain heuristic
            for _solver in solver.control.configuration.solver:
                _solver.heuristic="Domain"
            # find PREF_ATOM and SHOWN_ATOM options, and update domain flags
            for sign, atom_type, value, modifier in solver.options.on_opt_heur:
                if atom_type == PREF_ATOM:
                    self.pref = True
                    self.solver.set_holds_domain = True
                elif atom_type == SHOWN_ATOM:
                    self.shown = True
                    self.solver.set_shown_domain = True

    def on(self):
        return self.heuristic

    def create_shown_heuristic_rules(self):
        # gather signatures
        sigs = set()
        for atom in self.solver.shown_domain:
            sigs.add((atom.negative, atom.name, len(atom.arguments)))
        sigs = sorted(list(sigs))
        # for every signature, add heuristics and externals
        for negative, name, arity in sigs:
            atom = ("-" if negative else "") + name
            if arity > 0:
                variables = ",".join(["X" + str(i) for i in range(arity)])
                atom += "(" + variables + ")"
            rule = HEUR_RULE.replace("##" + HOLDS + "(X,0)", atom)
            rule = rule.replace(LAST_HOLDS, LAST_SHOWN)
            rule = rule.replace("(X).", "(" + atom + ").")
            self.shown_rule += rule

    def unsat(self):

        # return if no heuristics
        if not self.heuristic:
            return
        solver = self.solver

        # ground if not grounded
        if not self.grounded:
            self.grounded = True
            # if needed: create shown heuristic rules
            if self.shown == True:
                self.create_shown_heuristic_rules()
            # create heuristic program
            program = ""
            ## iterate over options
            for sign, atom_type, value, modifier in solver.options.on_opt_heur:
                rule = HEUR_RULE
                if atom_type == SHOWN_ATOM:
                    rule = self.shown_rule
                if sign == POS:
                    rule = rule.replace(": not ", ": ")
                rule = rule.replace("#v", value)
                rule = rule.replace("#m", modifier)
                program += rule
            ## add externals
            if self.pref:
                program += HEUR_EXTERNAL_HOLDS
            if self.shown:
                program += HEUR_EXTERNAL_SHOWN
            ## replace ##
            program = program.replace("##", solver.underscores)
            # add and ground it
            solver.add_and_ground(ON_OPT_HEUR_PROGRAM, [], program, [])

        # assign externals
        if self.pref:
            solver.assign_heuristic_externals(
                solver.get_holds_domain(),
                set(solver.get_holds()),
                solver.underscores + LAST_HOLDS
            )
        if self.shown:
            solver.assign_heuristic_externals(
                solver.get_shown_domain(),
                set(solver.get_shown()),
                solver.underscores + LAST_SHOWN
            )



# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

#!/usr/bin/python

from .ply import lex
from ..utils import utils
from ..utils import printer

#
# DEFINES
#

BASE  = utils.BASE
EMPTY = utils.EMPTY
ERROR_LEXER="lexer error, unexpected {}\n"


#
# Lexer
#

class Lexer(object):


    def __init__(self, underscores, options):
        self.__underscores = underscores
        self.__options     = options
        self.__show        = set()
        self.__error       = False
        self.lexer = lex.lex(module=self)
        self.lexer.push_state('normal')


    def new_file(self, filename):
        self.__filename   = filename
        self.__program    = (BASE, EMPTY)
        self.__code_start = 0
        self.__bc         = 0
        self.__eof        = False
        self.lexer.lineno = 1
        while self.lexer.lexstate != 'normal':
            self.lexer.pop_state()


    #
    # GET & SET
    #

    def get_underscores(self):
        return self.__underscores
    
    def get_error(self):
        return self.__error
    
    def get_show(self):
        return self.__show
    
    def set_program(self, program):
        self.__program = program

    def set_code_start(self, code_start):
        self.__code_start = code_start


    #
    # AUXILIARY FUNCTIONS
    #

    def __update_underscores(self, t):
        if t.value[0] == "_":
            i = 0
            while t.value[i] == "_": 
                i += 1
            if i>self.__underscores: 
                self.__underscores = i

    def __print_error(self, string, lexpos):
        col_ini = lexpos-self.lexer.lexdata.rfind('\n', 0, lexpos)
        loc = utils.Location(self.__filename, self.lexer.lineno,
                             col_ini, self.lexer.lineno, col_ini+len(string))
        printer.Printer().print_error_location(loc,ERROR_LEXER.format(string))


    def __eof_error(self, t):
        self.__error = True
        t.lexer.skip(1)
        self.__print_error("<EOF>", t.lexpos)

    def __eof_ok(self, t):
        if self.__eof:   # the second time, returns None 
            return None
        t.value        = t.lexer.lexdata[self.__code_start:t.lexer.lexpos]
        t.type         = 'CODE'
        t.lexer.lexpos = t.lexpos
        self.__eof     = True
        return t       # the first time, returns CODE


    #
    # START OF LEXING RULES
    #

    # Declare the state
    states = (
          ('comment',     'exclusive'),
          ('blockcomment','exclusive'),
          ('script',      'exclusive'),
          ('normal',      'exclusive'),
    )

    # List of token names. This is always required
    tokens = (
        'CODE', # CODE is only used by the defined states
                # the rest are only used in the INITIAL state
        'DOT',
        'PREFERENCE',
        'OPTIMIZE',
        'PROGRAM',
        'CONST',
        'INCLUDE',
        'LPAREN',
        'RPAREN',
        'LBRACE',
        'RBRACE',
        'SEM',
        'TWO_COLON',
        'COLON',
        'COMMA',
        'IDENTIFIER',
        'COND',
        'GTGT',
        'GT',
        'NOT',
        'INFIMUM',
        'SUPREMUM',
        'ANONYMOUS',
        'NUMBER',
        'VARIABLE',
        'STRING',
        'TRUE',
        'FALSE',
        'DOTS',
        'VBAR',
        'ADD',
        'SUB',
        'POW',
        'POW_NO_WS',
        'MOD',
        'MUL',
        'LT',
        'GEQ',
        'LEQ',
        'NEQ',
        'EQ',
        'SLASH',
        'AT',
        'AND',
        'XOR',
        'BNOT',
        'QUESTION',
        'MINIMIZE',
        'MAXIMIZE',
        'NEVER', # never reachable token
        'IF',    # added to expressions with NEVER to avoid warning
    )

    # Regular expression rules for simple tokens
    t_PREFERENCE = r'\#preference'
    t_OPTIMIZE   = r'\#optimize'
    t_PROGRAM    = r'\#program'
    t_CONST      = r'\#const'
    t_INCLUDE    = r'\#include'
    t_DOT        = r'\.'
    t_LPAREN     = r'\('
    t_RPAREN     = r'\)'
    t_LBRACE     = r'\{'
    t_RBRACE     = r'\}'
    t_SEM        = r';'
    t_TWO_COLON  = r'::'
    t_IF         = r':-'
    t_COLON      = r':'
    t_COMMA      = r','
    t_COND       = r'\|\|'
    t_GT         = r'>'
    t_INFIMUM   = r'\#inf(imum)?'
    t_SUPREMUM  = r'\#sup(remum)?'
    t_ANONYMOUS = r'_'
    t_NUMBER    = r'0|([1-9][0-9]*)'
    t_VARIABLE  = r'_*[A-Z][\'A-Za-z0-9_]*'
    t_STRING    = r'\" ( [^\\"\n] | (\\\") | (\\\\) | (\\n) )* \" '
    t_TRUE	    = r'\#true'
    t_FALSE	    = r'\#false'
    t_DOTS	    = r'\.\.'
    t_VBAR	    = r'\|'
    t_ADD	    = r'\+'
    t_SUB	    = r'\-'
    #t_POW	    = r'\*\*'
    t_MOD	    = r'\\'
    t_MUL	    = r'\*'
    t_GTGT	    = r'>>'
    t_LT	    = r'<'
    t_GEQ	    = r'>='
    t_LEQ	    = r'<='
    t_EQ	    = r'(==)|(=)'
    t_NEQ	    = r'(\!=)|(<>)'
    t_SLASH	    = r'\/'
    t_AT	    = r'\@'
    t_AND	    = r'\&'
    t_XOR	    = r'\^'
    t_BNOT	    = r'\~'
    t_QUESTION	= r'\?'
    t_MINIMIZE = r'\#minimize'
    t_MAXIMIZE = r'\#maximize'


    #
    # INITIAL state
    #

    def t_NOT(self, t):
        r'not'
        return t

    def t_WS(self, t):
        r'[\t\r ]+'
        pass

    def t_NL(self, t):
        r'\n+'
        self.lexer.lineno  += len(t.value)

    def t_POW(self, t):
        r'\*\*'
        if len(t.lexer.lexdata) > t.lexer.lexpos:
            if t.lexer.lexdata[t.lexer.lexpos] not in { '\t', '\r', ' ', '\n' }:
                t.type = "POW_NO_WS"
        return t

    def t_BLOCKCOMMENT(self, t):
        r'%\*'
        self.__bc = 1
        t.lexer.push_state('blockcomment')

    def t_COMMENT(self, t):
        r'%|(\#!)'
        t.lexer.push_state('comment')

    def t_IDENTIFIER(self, t):
        r'_*[a-z][\'A-Za-z0-9_]*'
        self.__update_underscores(t)
        return t

    def t_eof(self, t):
        self.__eof_error(t)
        return None


    #
    # normal state
    #

    def t_normal_eof(self, t):
        return self.__eof_ok(t)

    # strings: pass
    def t_normal_STRING(self, t):
        r'\" ( [^\\"\n] | (\\\") | (\\\\) | (\\n) )* \" '
        pass

    # identifier: update underscores'
    def t_normal_IDENTIFIER(self, t):
        r'_*[a-z][\'A-Za-z0-9_]*'
        self.__update_underscores(t)

    # show: update self.__show
    def t_normal_SHOW(self, t):
        r'\#show [\n\t\r ]* ([-\$]?_*[a-z][\'A-Za-z0-9_]* [\n\t\r ]* / [\n\t\r ]* (0|([1-9][0-9]*)))? [\n\t\r ]* \.'
        t.lexer.lexpos = t.lexpos + 1
        self.__show.add(self.__program)

    # project: return error
    def t_normal_PROJECT(self, t):
        r'\#project'
        self.__error = True
        t.lexer.skip(8)
        self.__print_error(t.value, t.lexpos)

    # optimization: in base return error, else continue
    def t_normal_OPTIMIZATION(self, t):
        r'(\#minimize)|(\#maximize)|(:~)'
        if self.__program == (BASE, EMPTY):
            self.__error = True
            t.lexer.skip(len(t.value))
            self.__print_error(t.value, t.lexpos)


    #
    # changing state:
    #  %*                   -> blockcomment
    #  % or \#!             ->      comment
    #  #script (python|lua) ->       script
    #  #directive           ->      INITIAL
    #

    def t_normal_BLOCKCOMMENT(self, t):
        r'%\*'
        self.__bc = 1
        t.lexer.push_state('blockcomment')

    def t_normal_COMMENT(self, t):
        r'%|(\#!)'
        t.lexer.push_state('comment')

    def t_normal_SCRIPT(self, t):
        r'\#script[\t\r ]*\([\t\r ]*(python|lua)[\t\r ]*\)'
        t.lexer.push_state('script')
        t.lexer.lexpos = t.lexpos

    # push INITIAL state, reset lexer lexpos, and return CODE
    def t_normal_DIRECTIVE(self, t):
        r'(\#preference)|(\#optimize)|(\#program)|(\#const)|(\#include)'
        #r'(\#preference)|(\#optimize)|(\#program)|(\#const)|(\#include)|(\#minimize)|(\#maximize)'
        #if ((t.value == "#maximize" or t.value == "#minimize") and
        #    ((self.__program != (BASE, EMPTY)) or
        #     not self.__options['minimize'])):
        #    t.lexer.lexpos = t.lexpos + 1
        #    return
        t.lexer.push_state('INITIAL')
        t.value = t.lexer.lexdata[self.__code_start:t.lexpos]
        t.type = 'CODE'
        t.lexer.lexpos = t.lexpos
        return t

    def t_normal_NL(self, t):
        r'\n'
        self.lexer.lineno += 1

    def t_normal_ANY(self, t):
        r'[\000-\377]'
        pass


    #
    # blockcomment state
    #

    def t_blockcomment_eof(self, t):
        self.__eof_error(t)
        return None
    
    def t_blockcomment_ENDBLOCKCOMMENT(self, t):
        r'\*%'
        self.__bc -= 1
        if self.__bc == 0: 
            t.lexer.pop_state()

    def t_blockcomment_BLOCKCOMMENT(self, t):
        r'%\*'
        self.__bc += 1

    def t_blockcomment_COMMENT(self, t):
        r'%'
        t.lexer.push_state('comment')

    def t_blockcomment_NL(self, t):
        r'\n'
        self.lexer.lineno += 1

    def t_blockcomment_ANY(self, t):
        r'[\000-\377]'
        pass


    #
    # comment state
    #

    def t_comment_eof(self, t):
        if self.__bc > 0:
            self.__eof_error(t)
            return None 
        return self.__eof_ok(t)

    def t_comment_NL(self, t):
        r'\n'
        t.lexer.lineno += 1
        t.lexer.pop_state()

    def t_comment_ANY(self, t):
        r'[\000-\377]'
        pass


    #
    # script state
    #

    def t_script_eof(self, t):
        self.__eof_error(t)
        return None

    def t_script_END(self, t):
        r'\#end'
        t.lexer.pop_state()

    def t_script_NL(self, t):
        r'\n'
        self.lexer.lineno += 1

    def t_script_ANY(self, t):
        r'[\000-\377]'
        pass


    #
    # error handling
    #

    def t_ANY_error(self, t):
        self.__error = True
        t.lexer.skip(1)
        self.__print_error(t.value[0],t.lexpos)




# MIT License
#
# Copyright (c) 2017 Javier Romero
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

#
# IMPORTS
#

import clingo
import sys
import math
from threading import Condition
from . import controller
from ..utils import printer
from ..utils import utils
from .metasp import metasp

# TIMING
import time
times = {}
FREQUENCY = 1
def start_clock(clock):
    times[clock] = time.clock()
def check_clock(clock):
    return time.clock() - times[clock]


#
# DEFINES
#

# nodes and control
START         = "START"
START_LOOP    = "START_LOOP"
SOLVE         = "SOLVE"
SAT           = "SAT"
UNSAT         = "UNSAT"
UNKNOWN       = "UNKNOWN"
END_LOOP      = "END_LOOP"
END           = "END"
SATISFIABLE   = utils.SATISFIABLE                   # used also by controller
UNSATISFIABLE = "UNSATISFIABLE"

# for meta-programming
META_SIMPLE    = utils.META_SIMPLE
META_COMBINE   = utils.META_COMBINE
METAPROGRAM    = utils.METAPROGRAM
QUERY          = utils.QUERY
QUERY_PROGRAM  = utils.QUERY_PROGRAM
METAUNSAT      = utils.METAUNSAT
METAUNSAT_BASE = utils.METAUNSAT_BASE

# strings
STR_ANSWER             = "Answer: {}"
STR_OPTIMUM_FOUND      = "OPTIMUM FOUND"
STR_OPTIMUM_FOUND_STAR = "OPTIMUM FOUND *"
STR_MODEL_FOUND        = utils.STR_MODEL_FOUND      # used also by controller
STR_MODEL_FOUND_STAR   = utils.STR_MODEL_FOUND_STAR # used also by controller
STR_UNSATISFIABLE      = "UNSATISFIABLE"
STR_SATISFIABLE        = "SATISFIABLE"
STR_LIMIT              = "MODEL FOUND (SEARCH LIMIT)"
STR_BENCHMARK_CLOCK    = "BENCHMARK"
STR_BENCHMARK_FILE     = "benchmark.txt"

# program names
DO_HOLDS = "do_holds"
DO_HOLDS_APPROX = "do_holds_approx"
DO_HOLDS_DELETE_BETTER = "do_holds_delete_better"
DO_HOLDS_AT_ZERO = "do_holds_at_zero"
OPEN_HOLDS = "open_holds"
VOLATILE_FACT = "volatile_fact"
VOLATILE_EXT = "volatile_external"
DELETE_MODEL = "delete_model"
DELETE_MODEL_VOLATILE = "delete_model_volatile"
DELETE_MODEL_APPROX = "delete_model_approx"
UNSAT_PRG = "unsat"
NOT_UNSAT_PRG = "not_unsat"
CMD_HEURISTIC = "cmd_heuristic"
PROJECT_CLINGO = "project_clingo"
PREFP = utils.PREFP
PBASE = utils.PBASE
APPROX = utils.APPROX
HEURISTIC = utils.HEURISTIC
UNSATP = utils.UNSATP
UNSATPBASE = utils.UNSATPBASE

# predicate and term names
VOLATILE      = utils.VOLATILE
MODEL         = utils.MODEL
HOLDS         = utils.HOLDS
UNSAT_ATOM    = utils.UNSAT
PREFERENCE    = utils.PREFERENCE
HOLDS_AT_ZERO = utils.HOLDS_AT_ZERO
CSP           = "$"
MODEL_DELETE_BETTER = clingo.parse_term("delete_better")
DELETE_MODEL_VOLATILE_ATOM = "delete_model_volatile_atom"

# messages
WRONG_APPEND = """\
incorrect input to @append function, this should not happen"""
SAME_MODEL = """\
same stable model computed twice, there is an error in the input, \
probably an incorrect preference program"""
WARNING_NO_OPTIMIZE = """WARNING: no optimize statement, \
computing non optimal stable models"""
STR_BETTER_THAN_UNKNOWN = "BETTER THAN MODEL(S): {}"
STR_UNKNOWN_OPTIMAL = """\nINFO: The following MODEL(S) FOUND (with SEARCH LIMIT) \
are OPTIMAL MODEL(S): {}"""
STR_UNKNOWN_NONOPTIMAL = """\nINFO: All MODEL(S) FOUND (with SEARCH LIMIT) \
*could* be OPTIMAL MODEL(S): {}"""
STR_BENCHMARK_MSG = """Solving Time {}: {:.2f}s"""

#
# AUXILIARY PROGRAMS
#

TOKEN    = "##"
PROGRAMS = \
  [(DO_HOLDS_AT_ZERO,       [],"""
#show ##holds_at_zero(X) : ##""" + HOLDS + """(X,0)."""),
   (DO_HOLDS,            ["m"],"""
##""" + HOLDS + """(X,m) :- X = @get_holds()."""),
   (OPEN_HOLDS,          ["m"],"""
{ ##""" + HOLDS + """(X,m) } :- X = @get_holds_domain()."""),
   (VOLATILE_FACT, ["m1","m2"],"""
##""" + VOLATILE + """(##m(m1),##m(m2))."""),
   (VOLATILE_EXT,  ["m1","m2"],"""
#external ##""" + VOLATILE + """(##m(m1),##m(m2))."""),
   (DELETE_MODEL,           [],"""
:-     ##""" + HOLDS + """(X,0) : X = @get_holds();
   not ##""" + HOLDS + """(X,0) : X = @get_nholds()."""),
   (DELETE_MODEL_VOLATILE,           ["m"],"""
#external ##""" + DELETE_MODEL_VOLATILE_ATOM + """(m).
:-     ##""" + HOLDS + """(X,0) : X = @get_holds();
   not ##""" + HOLDS + """(X,0) : X = @get_nholds();
   not ##""" + DELETE_MODEL_VOLATILE_ATOM + """(m)."""),
   (UNSAT_PRG,     ["m1","m2"],"""
:- not ##""" + UNSAT_ATOM + """(##m(m1),##m(m2)),
       ##""" +   VOLATILE + """(##m(m1),##m(m2))."""),
   (NOT_UNSAT_PRG, ["m1","m2"],"""
:-     ##""" + UNSAT_ATOM + """(##m(m1),##m(m2)),
       ##""" +   VOLATILE + """(##m(m1),##m(m2))."""),
   (CMD_HEURISTIC,   ["v","m"],"""
#heuristic ##""" + HOLDS + """(X,0) : ##""" + PREFERENCE +
    """(_,_,_,for(X),_). [v@0,m]"""),
   (DO_HOLDS_DELETE_BETTER,        [],"""
##""" + HOLDS + """(X,""" + str(MODEL_DELETE_BETTER) + """) :- ##""" +
    HOLDS + """(X,0)."""),
   (PROJECT_CLINGO,           [],"""
#project  ##""" + HOLDS + """/2."""),
  ]
PROGRAMS_APPROX = \
  [(DO_HOLDS_APPROX,            ["m","mm"],"""
##""" + HOLDS + """(X,m) :- X = @get_holds_approx(mm)."""),
   (DELETE_MODEL_APPROX,           ["mm"],"""
:-     ##""" + HOLDS + """(X,0) : X = @get_holds_approx(mm);
   not ##""" + HOLDS + """(X,0) : X = @get_nholds_approx(mm)."""),
  ]
UNSAT_PREFP  = (PREFP,  ["m1","m2"], "##" + UNSAT_ATOM +"(##m(m1),##m(m2)).")

#
# Auxiliary Classes (EndException, Options)
#

class EndException(Exception):
    pass

class Options:
    pass

#
# Solver
#

class Solver:

    def __init__(self, control, options, control_proxy, observer):
        # control and options
        self.control           = control
        self.options           = Options()
        for key, value in options.items():
            setattr(self.options, key, value)
        self.control_proxy     = control_proxy
        self.observer = observer
        # strings
        self.underscores       = utils.underscores
        self.volatile_str      = self.underscores + VOLATILE
        self.model_str         = self.underscores + MODEL
        self.holds_at_zero_str = self.underscores + HOLDS_AT_ZERO
        self.holds_str         = self.underscores + HOLDS
        self.unsat_str         = self.underscores + UNSAT_ATOM
        self.delete_str        = self.underscores + DELETE_MODEL_VOLATILE_ATOM
        # holds
        self.holds             = []
        self.nholds            = []
        # others
        self.step = 1
        self.last_unsat = True
        self.opt_models = 0
        self.models = 0
        self.more_models = True
        self.old_holds = None
        self.shown = []
        self.solving_result = None
        self.externals  = dict()
        self.improving  = []
        self.not_improving  = []
        self.store_holds = True
        self.store_nholds = False
        self.approx_opt_models = []
        self.assumptions = []
        self.last_model = None
        self.sequences = {}
        self.unknown = []
        self.unknown_non_optimal = []
        self.grounded_delete_better = False
        self.mapping = {}
        self.unsat_program = PREFP
        self.unsat_program_base = None
        # for weak mode
        self.control.configuration.solve.opt_mode = 'ignore' # by default ignore
        self.optN = False
        self.on_optimal = None
        # holds and shown domains
        self.set_holds_domain = False
        self.holds_domain = []
        self.set_shown_domain = False
        self.shown_domain = []
        # exiting
        self.exited = False
        # functions
        self.same_shown_function = self.same_shown_underscores
        # strings
        self.str_found      = STR_OPTIMUM_FOUND
        self.str_found_star = STR_OPTIMUM_FOUND_STAR
        # printer
        self.printer = printer.Printer()
        #if self.options.max_models == 1 and not self.options.improve_limit:
        #    self.store_nholds = False
        self.saved_stats = False
        if self.options.benchmark:
            start_clock(STR_BENCHMARK_CLOCK)
            with open(STR_BENCHMARK_FILE, 'w') as f:
                f.write("")

    #
    # AUXILIARY
    #

    def get_external(self, m1, m2):
        external = self.externals.get((m1,m2))
        if external is None:
            f1 = clingo.Function(self.model_str, [m1])
            f2 = clingo.Function(self.model_str, [m2])
            external = clingo.Function(self.volatile_str, [f1, f2])
            self.externals[(m1,m2)] = external
        return external

    def get_holds(self):
        return self.holds

    def get_holds_domain(self):
        return self.holds_domain

    def do_set_holds_domain(self):
        self.holds_domain = [
            i.symbol.arguments[0] for i in
                self.control.symbolic_atoms.by_signature(self.holds_str, 2)
                if str(i.symbol.arguments[1]) == "0"
        ]

    def get_nholds(self):
        return self.nholds

    def get_holds_function(self, term, y):
        return clingo.Function(self.holds_str, [term, clingo.Number(y)])

    def get_unsat_function(self, term, y):
        return clingo.parse_term("{}({}{}({}),{}{}({}))".format(
            self.unsat_str,
            self.underscores, MODEL, str(term),
            self.underscores, MODEL, y
            ))

    def cat(self, tuple):
        if tuple.arguments:
            return "".join([str(i) for i in tuple.arguments]).replace('"',"")
        else:
            return str(tuple)

    def append(self, elem, alist):
        if alist.name == "" and len(alist.arguments):
            return clingo.Function("", [elem] + alist.arguments)
        else:
            return clingo.Function("", [elem] + [alist])

    def save_stats(self):
        if not self.saved_stats or check_clock(STR_BENCHMARK_CLOCK) > FREQUENCY:
            self.saved_stats = True
            with open(STR_BENCHMARK_FILE, 'w') as f:
                self.print_stats(file=f)
            start_clock(STR_BENCHMARK_CLOCK)


    #
    # USED BY ASPRIN LIBRARY (to be copied at metasp/metasp.py)
    #

    def exp2(self, x):
        return int(math.pow(2,x.number))

    def get(self, atuple, index):
        try:
            return atuple.arguments[index.number]
        except:
            return atuple

    def get_mode(self):
        return self.options.solving_mode

    def get_sequence(self, name, elem):
        string = str(name)
        if string in self.sequences:
            self.sequences[string] += 1
        else:
            self.sequences[string]  = 1
        return self.sequences[string]

    def length(self, atuple):
        try:
            return len(atuple.arguments)
        except:
            return 1

    def log2up(self, x):
        return int(math.ceil(math.log(x.number,2)))

    #
    # CLINGO PROXY
    #

    def add_and_ground(self, name, params, program, params_instances):
        self.control.add(name, params, program)
        self.ground([(name, params_instances)], self)

    def add_encodings(self):
        for i in PROGRAMS:
            self.control.add(i[0], i[1], i[2].replace(TOKEN, self.underscores))
        self.ground([(DO_HOLDS_AT_ZERO, [])], self)

    def add_projection(self):
        self.ground([(PROJECT_CLINGO, [])], self)
        self.control.configuration.solve.project = 'project'

    def add_unsat_to_preference_program(self):
        self.control.add(UNSAT_PREFP[0], UNSAT_PREFP[1],
                         UNSAT_PREFP[2].replace(TOKEN, self.underscores))

    def check_errors(self):
        pr, control, u = self.printer, self.control, self.underscores
        error = False
        for atom in control.symbolic_atoms.by_signature(
            u + "_" + utils.WARN_PRED, 1
        ):
            string = self.cat(atom.symbol.arguments[0]) + "\n"
            pr.print_spec_warning(string)
        for atom in control.symbolic_atoms.by_signature(
            u + "_" + utils.ERROR_PRED, 1
        ):
            string = self.cat(atom.symbol.arguments[0]) + "\n"
            pr.print_spec_error(string)
            error = True
        if error:
            raise Exception("parsing failed")

    def check_last_model(self):
        if self.old_holds  == self.holds:
            self.printer.do_print()
            raise Exception(SAME_MODEL)
        self.old_holds  = self.holds

    def clean_up(self):
        self.control.cleanup()

    def enumerate_on_model(self, model):
        true = model.symbols(shown=True)
        self.shown = [i for i in true if i.name != self.holds_at_zero_str]
        # we may enumerate one more model than what is needed,
        # so we check whether we already computed all models
        if self.options.max_models != 0 and \
           self.options.max_models == self.opt_models:
            return
        # self.same_shown_function is modified by EnumerationController
        # at controller.py
        if self.enumerate_flag or not self.same_shown_function():
            self.models     += 1
            self.opt_models += 1
            if self.options.quiet in {0,1}:
                self.print_answer()
            else:
                self.print_str_answer()
            self.print_optimum_string(True)

    def enumerate(self, add_one=True):
        # models
        control, solve_conf = self.control, self.control.configuration.solve
        old_models = solve_conf.models
        self.set_control_models()
        # we add one if not computing all and add_one (by default)
        if solve_conf.models != "0" and add_one:
            solve_conf.models = str(int(solve_conf.models) + 1)
        # assumptions
        ass  = [ (self.get_holds_function(x,0),  True) for x in self.holds ]
        ass += [ (self.get_holds_function(x,0), False) for x in self.nholds]
        # solve
        self.old_shown, self.enumerate_flag = self.shown, False
        self.solve(assumptions = ass + self.assumptions,
                   on_model = self.enumerate_on_model)
        self.shown = self.old_shown
        solve_conf.models = old_models

    def get_preference_parts(self, x, y, better, volatile):
        if better:
            parts = [(             PREFP, [x, y]), (NOT_UNSAT_PRG, [x,y])]
        else:
            parts = [(self.unsat_program, [x, y]), (    UNSAT_PRG, [x,y])]
        if volatile:
            parts.append((VOLATILE_EXT,  [x,y]))
        else:
            parts.append((VOLATILE_FACT, [x,y]))
        return parts

    def get_shown(self):
        return self.shown

    def get_shown_domain(self):
        return self.shown_domain

    def ground(self, *args):
        self.control_proxy.ground(*args)

    def ground_cmd_heuristic(self):
        params = [clingo.parse_term(i) for i in self.options.cmd_heuristic]
        self.ground([(CMD_HEURISTIC, params)], self)

    def ground_heuristic(self):
        self.ground([(HEURISTIC, [])], self)

    def ground_holds(self, step):
        self.ground([(DO_HOLDS, [step])], self)

    def ground_holds_delete_better(self):
        if not self.grounded_delete_better:
            self.ground([(DO_HOLDS_DELETE_BETTER, [])], self)
            self.grounded_delete_better = True

    def ground_preference_base(self):
        self.ground([(PBASE, [])], self)

    def ground_preference_program(self, volatile):
        control, prev_step = self.control, self.step-1
        parts = self.get_preference_parts(0, prev_step, True, volatile)
        control.ground(parts, self)
        if volatile:
            if self.options.release_last:
                self.relax_previous_models()
            control.assign_external(self.get_external(0,prev_step), True)
            self.improving.append(prev_step)

    def ground_unsatp_base(self):
        self.ground([(self.unsat_program_base, [])], self)

    def handle_optimal_model(self, step, delete_model_volatile,
                             delete_worse, delete_better, volatile):
        if not delete_model_volatile:
            parts = [(DELETE_MODEL, [])]
        else:
            parts = [(DELETE_MODEL_VOLATILE, [step])]
        if delete_worse:
            parts += self.get_preference_parts(step, 0, False, volatile)
        # TODO: In base setting, use same preference program as for improving
        if delete_better:
            parts += self.get_preference_parts(
                MODEL_DELETE_BETTER, step, False, volatile
            )
        self.ground(parts, self)
        if volatile:
            self.handle_volatile_optimal_model(step, delete_worse, delete_better)

    def handle_volatile_optimal_model(self, step, delete_worse, delete_better):
        if delete_worse:
            self.not_improving.append((step,0))
        if delete_better:
            self.not_improving.append((MODEL_DELETE_BETTER,step))
        for x,y in self.not_improving:        #activate
            self.control.assign_external(self.get_external(x,y),True)
        if not self.options.no_opt_improving: #reset
            self.not_improving = []

    def no_optimize(self):
        optimize = self.underscores + utils.OPTIMIZE
        for atom in self.control.symbolic_atoms.by_signature(optimize, 1):
            return False
        return True

    def on_model(self, model):
        self.holds, self.nholds, self.shown = [], [], []
        if self.set_shown_domain:
            self.shown_domain_set(model)
        for a in model.symbols(shown=True):
            if a.name != self.holds_at_zero_str:
                self.shown.append(a)
            elif self.store_holds:
                self.holds.append(a.arguments[0])
        if self.store_nholds:
            self.nholds = [
                x for x in self.holds_domain if x not in set(self.holds)
            ]

    def on_model_single(self, model):
        # call on_model
        self.on_model(model)
        # update numbers and print
        self.models += 1
        self.opt_models += 1
        self.print_str_answer()
        if self.options.quiet in (0, 1):
            self.print_shown()
        self.print_optimum_string()
        if not self.keep_shown:
            self.shown = []

    def set_config(self):
        try:
            self.iconfigs = (self.iconfigs + 1) % len(self.options.configs)
        except:
            self.iconfigs = 0
        self.control.configuration.configuration = self.options.configs[self.iconfigs]

    def set_control_models(self):
        solve_conf = self.control.configuration.solve
        if self.options.max_models == 0:
            solve_conf.models = "0"
        else:
            solve_conf.models = str(self.options.max_models - self.opt_models)

    def set_solving_result(self, result):
        # set solving result
        if result.satisfiable:
            self.solving_result = SATISFIABLE
        elif result.unsatisfiable:
            self.solving_result = UNSATISFIABLE
        else:
            self.solving_result = UNKNOWN

    # TODO: do not allow --preference-unsat and meta?
    def set_unsat_program(self):
        # auxiliary
        def set_and_ground(base, incremental):
            self.unsat_program_base = base
            self.unsat_program = incremental
            self.ground([(self.unsat_program_base, [])], self)
        # meta
        if self.options.meta == META_COMBINE:
            # meta_incremental() creates and adds METAUNSAT_BASE and METAUNSAT
            self.meta_incremental()
            set_and_ground(METAUNSAT_BASE, METAUNSAT)
        # preference_unsat
        if self.options.preference_unsat:
            set_and_ground(UNSATPBASE, UNSATP)

    def shown_domain_append(self, atom):
        if atom.type == clingo.SymbolType.Function and len(atom.name) > 0:
            self.shown_domain.append(atom)
            #name = ("-" if atom.negative else "") + atom.name
            #self.shown_domain.add((name, len(atom.arguments)))

    def shown_domain_set(self, model):
        self.set_shown_domain = False
        # gather all true and false atoms
        tatoms = model.symbols(atoms=True)
        fatoms = model.symbols(atoms=True, complement=True)
        # add shown elements that are true and false atoms
        for atom in model.symbols(shown=True):
            if atom in tatoms:
                self.shown_domain_append(atom)
        for atom in model.symbols(shown=True, complement=True):
            if atom in fatoms:
                self.shown_domain_append(atom)

    def solve(self, *args, **kwargs):
        if self.options.configs is not None:
            self.set_config()
        result = self.control_proxy.solve(*args, **kwargs)
        self.set_solving_result(result)
        return result

    def solve_heuristic(self):
        h = []
        # gather heuristics
        for _solver in self.control.configuration.solver:
            h.append(_solver.heuristic)
            _solver.heuristic="Domain"
        # solve
        self.solve(assumptions=self.assumptions, on_model=self.on_model)
        # restore heuristics
        i = 0
        for _solver in self.control.configuration.solver:
            _solver.heuristic = h[i]
            i += 1

    def solve_single_on_optimal(self):
        self.control.configuration.solve.models = 1
        self.store_holds, self.store_nholds, self.keep_shown = [True]*3
        if not self.set_holds_domain: # required to store_nholds
            raise utils.FatalException() 
        self.printer.do_print("Solving...")
        while True:
            result = self.solve(on_model=self.on_model_single)
            # unsat
            if self.opt_models == 0:
                self.print_unsat()
                break
            # search space exhausted
            if result.exhausted:
                self.more_models = False
                break
            # computed all models
            if self.options.max_models == self.opt_models:
                break
            # enumerate if no projection
            if not self.options.project:
                self.enumerate() # TODO: prepare for this!
            # execute on_optimal
            self.on_optimal.unsat()
            # delete model
            self.ground([(DELETE_MODEL, [])], self) # TODO: store holds and nholds

    def solve_single(self):
        # if on_optimal, call and return
        if self.on_optimal.on():
            self.solve_single_on_optimal()
            return
        # add projection if needed
        if self.options.project:
            self.add_projection()
        # prepare to solve
        self.control.configuration.solve.models = self.options.max_models
        self.store_holds, self.store_nholds, self.keep_shown = [False]*3
        self.printer.do_print("Solving...")
        # solve and finish
        result = self.solve(on_model=self.on_model_single)
        if result.exhausted:
            self.more_models = False
        if self.opt_models == 0:
           self.print_unsat()

    def solve_unknown(self):
        self.solving_result = UNKNOWN

    def solve_unsat(self):
        self.solving_result = UNSATISFIABLE

    def symbol2str(self,symbol):
        if symbol.name == CSP:
            return str(symbol.arguments[0]) + "=" + str(symbol.arguments[1])
        return str(symbol)

    def print_answer(self):
        self.printer.do_print(STR_ANSWER.format(self.models))
        self.printer.do_print(" ".join(map(self.symbol2str, self.shown)))

    def print_better_than_unknown(self, unknowns, mapping):
        self.print_unknowns(STR_BETTER_THAN_UNKNOWN, unknowns, mapping)

    def print_limit_string(self):
        self.printer.do_print(STR_LIMIT)

    def print_no_optimize_warning(self):
        self.printer.print_warning(WARNING_NO_OPTIMIZE)

    def print_optimum_string(self, star=False):
        if not star:
            self.printer.do_print(self.str_found)
        else:
            self.printer.do_print(self.str_found_star)

    def print_shown(self):
        self.printer.do_print(" ".join(map(self.symbol2str, self.shown)))

    def print_steps_message(self):
        if self.opt_models == 0:
            self.printer.do_print(STR_SATISFIABLE)

    def print_str_answer(self):
        self.printer.do_print(STR_ANSWER.format(self.models))

    def print_unknowns(self, string, unknowns, mapping):
        if unknowns:
            self.printer.do_print(string.format(
                " ".join([str(mapping[i]) for i in unknowns])
            ))

    def print_unknown_nonoptimal_models(self, unknowns, mapping):
        self.print_unknowns(STR_UNKNOWN_NONOPTIMAL, unknowns, mapping)

    def print_unknown_optimal_models(self, unknowns, mapping):
        self.print_unknowns(STR_UNKNOWN_OPTIMAL, unknowns, mapping)

    def print_unsat(self):
        self.printer.do_print(STR_UNSATISFIABLE)

    def relax_previous_models(self):
        for i in self.improving:
            self.control.release_external(self.get_external(0, i))
        self.improving = []

    def relax_optimal_models(self):
        for x,y in self.not_improving:
            self.control.assign_external(self.get_external(x,y),False)

    def same_shown(self):
        if set(self.old_shown) == set(self.shown):
            self.enumerate_flag = True
            return True
        return False

    def same_shown_false(self):
        return False

    def same_shown_underscores(self):
        u = self.underscores
        s1 = set([i for i in self.old_shown if not str(i).startswith(u)])
        s2 = set([i for i in     self.shown if not str(i).startswith(u)])
        if s1 == s2:
            self.enumerate_flag = True
            return True
        return False

    #
    # on_opt_heur option
    #

    def assign_heuristic_externals(self, domain, atoms, external_name):
        for i in domain:
            external = clingo.Function(external_name, [i])
            if i in atoms:
                self.control.assign_external(external, True)
            else:
                self.control.assign_external(external, False)

    #
    # ground once method
    #

    def ground_open_preference_program(self):
        parts = self.get_preference_parts(0, -1, True, True)
        parts.append((OPEN_HOLDS, [-1]))
        self.ground(parts, self)

    def turn_off_preference_program(self):
        self.control.assign_external(self.get_external(0,-1), False)
        self.assumptions = (
            [(self.get_holds_function(i,-1),False) for i in  self.holds_domain]
        )

    def turn_on_preference_program(self):
        self.control.assign_external(self.get_external(0,-1), True)
        self.assumptions = (
            [(self.get_holds_function(i,-1),True)  for i in  self.holds] +
            [(self.get_holds_function(i,-1),False) for i in self.nholds]
        )

    #
    # weak approximation
    #

    def on_model_approx(self, model):
        # for all models
        self.models += 1
        self.print_str_answer()
        # for models not proved to be optimal
        if not model.optimality_proven and model.cost!=[]:
            if self.options.quiet == 0:
                self.on_model(model)
                self.print_shown()
            return
        # for models proved to be optimal
        self.on_model(model)
        if self.options.quiet in {0,1}:
            self.print_shown()
        self.approx_opt_models.append(self.holds)
        self.opt_models += 1
        self.print_optimum_string()

    def get_holds_approx(self, i):
        return self.approx_opt_models[int(str(i))]

    def get_nholds_approx(self, i):
        return [
            x for x in self.holds_domain
            if x not in set(self.approx_opt_models[int(str(i))])
        ]

    def solve_approx(self):
        # approximation programs
        self.ground([(APPROX, [])], self)
        for i in PROGRAMS_APPROX:
            self.control.add(i[0], i[1], i[2].replace(TOKEN, self.underscores))
        # is on on_optimal?
        on_on_optimal = self.on_optimal.on()
        # set opt_mode
        self.control.configuration.solve.opt_mode = 'optN'
        # project
        if self.options.project and not on_on_optimal:
            self.add_projection()
        # loop
        self.printer.do_print("Solving...")
        first = True
        while True:
            # solve
            prev_opt_models, self.approx_opt_models = self.opt_models, [[]]
            self.set_control_models()
            if on_on_optimal: # go one by one
                self.control.configuration.solve.models = 1
            self.solve(on_model=self.on_model_approx)
            satisfiable = self.solving_result == SATISFIABLE
            # break if unsat or computed all
            if not satisfiable or self.options.max_models == self.opt_models:
                break
            # if on on_optimal
            if on_on_optimal:
                if not self.options.project:
                    self.control.configuration.solve.opt_mode = 'ignore'
                    self.enumerate()
                    self.control.configuration.solve.opt_mode = 'optN'
                self.on_optimal.unsat()
            # if first, set unsat programs
            if first:
                self.set_unsat_program()
            # add programs
            parts = []
            for mm in range(1, len(self.approx_opt_models)):
                m = mm + prev_opt_models
                parts += [(DELETE_MODEL_APPROX, [mm]),
                          (DO_HOLDS_APPROX,   [m,mm])]
                if self.options.total_order and m>1:
                    continue
                parts += self.get_preference_parts(m, 0, False, False)
                if self.options.delete_better:
                    parts += self.get_preference_parts(0, m, False, False)
            self.ground(parts, self)
            # if on_optimal
        # end
        if self.opt_models == 0:
            self.print_unsat()
        self.more_models = True if satisfiable else False
        self.end()


    #
    # unknown (--improve-limit)
    #

    def computed_all(self):
        return (self.options.max_models != 0 and \
                self.opt_models == self.options.max_models)

    def enumerate_unknown(self):

        # if no unknowns, or computed all: return
        if not self.unknown:
            self.more_models = False
            return
        if self.computed_all():
            return
        # if quick: print message and update opt_models
        if self.options.improve_limit[3]:
            self.print_unknown_optimal_models(self.unknown, self.mapping)
            self.opt_models += len(self.unknown)
            return
        # if improve_no_check: print message
        if self.options.improve_limit[4]:
            self.print_unknown_nonoptimal_models(self.unknown, self.mapping)
            return

        # ELSE: print *shown* atoms
        # create boolean array representing unknown
        unknowns = [False] * (self.last_model + 1)
        for i in self.unknown:
            unknowns[i] = True
        # create holds dictionary for unknowns
        holds = {}
        for i in self.control.symbolic_atoms.by_signature(self.holds_str, 2):
            try:
                step = int(i.symbol.arguments[1].number)
                if unknowns[step]:
                    alist = holds.setdefault(step, [])
                    alist.append(i.symbol.arguments[0])
            except:
                pass

        # enumerate iterating over holds
        old = self.same_shown_function
        self.same_shown_function = self.same_shown_false
        for step in self.unknown:
            if self.computed_all():
                return
            # pre
            self.holds  = holds.get(step, [])
            self.nholds = [
                x for x in self.holds_domain if x not in set(self.holds)
            ]
            delete_model = clingo.parse_term(
                "{}({})".format(self.delete_str, step)
            )
            self.control.assign_external(delete_model, True)
            if self.options.project:
                old = self.options.max_models
                self.options.max_models = self.opt_models + 1
            # enumerate
            self.enumerate(add_one=False)
            # post
            if self.options.project:
                self.options.max_models = old
            self.control.release_external(delete_model)
        self.same_shown_function = old
        self.more_models = False

    def on_model_unknown(self, model):
        atoms = model.symbols(atoms=True)
        for i in self.unknown:
            if not self.get_unsat_function(MODEL_DELETE_BETTER, i) in atoms:
                self.unknown_non_optimal.append(i)

    def handle_unknown_models(self, result):

        # if improve_no_check, add to unknown list if unknown
        if self.options.improve_limit[4]:
            if result == UNKNOWN:
                self.unknown.append(self.last_model)
                self.mapping[self.last_model] = self.models
            return

        # else check if some unknowns are worse than the latest model
        # assumptions
        ass  = [ (self.get_holds_function(x,0),  True) for x in self.holds ]
        ass += [ (self.get_holds_function(x,0), False) for x in self.nholds]
        ass += [                             (x, True) for x in self.shown]
        # turn unknowns on
        for i in self.unknown:
            self.control.assign_external(
                self.get_external(MODEL_DELETE_BETTER, i), True
            )
        # solve
        self.unknown_non_optimal = []
        if self.unknown:
            self.solve(assumptions = ass + self.assumptions,
                       on_model = self.on_model_unknown)
        # print which unknown are not optimal
        self.print_better_than_unknown(self.unknown_non_optimal, self.mapping)
        # initialize array to update unknown
        update_unknown = [self.last_model] if result == UNKNOWN else []
        # release non optimal, and update unknown
        for i in self.unknown:
            if i in self.unknown_non_optimal:
                self.control.release_external(
                    self.get_external(MODEL_DELETE_BETTER, i)
                )
                self.control.release_external(
                    self.get_external(i, 0)
                )
                del self.mapping[i]
            else:
                self.control.assign_external(
                    self.get_external(MODEL_DELETE_BETTER, i), False
                )
                update_unknown.append(i)
        self.unknown = update_unknown
        # update not_improving
        self.not_improving = [
            x for x in self.not_improving
            if x not in set([(y,0) for y in self.unknown_non_optimal])
        ]
        # if UNKNOWN, add delete better for last model (w/out unsat constraint)
        if result == UNKNOWN:
            x, y  = MODEL_DELETE_BETTER, self.last_model
            parts = [(self.unsat_program, [x, y]), (VOLATILE_EXT,  [x,y])]
            self.ground(parts, self)
            #self.control.ground(parts, self)
            # also, add mapping
            self.mapping[self.last_model] = self.models
        #
        # leftover from merge
        #
        #self.not_improving.difference_update(
        #    [(x,0) for x in self.unknown_non_optimal]
        #)
        ## add delete better for last model (without unsat constraint)
        #x, y  = MODEL_DELETE_BETTER, self.last_model
        #parts = [(PREFP, [x, y]), (VOLATILE_EXT,  [x,y])]
        #self.ground(parts, self)

    #
    # meta-programming
    #

    def meta_simple(self):
        # choose meta implementation
        if self.options.meta_binary:
            meta = metasp.MetaspBinary(self)
        else:
            meta = metasp.MetaspPython(self)
        # get meta program
        meta_program = meta.get_meta_program()
        # add and ground
        self.control.add(METAPROGRAM, [], meta_program)
        self.ground([(METAPROGRAM, [])])
        # if query: adds the query and grounds
        if self.options.meta_query:
            qname, qprogram = QUERY, QUERY_PROGRAM
            self.control.add(qname, [], qprogram)
            self.ground([(qname, [])])
        # solve single
        self.solve_single()

    def meta_incremental(self):
        # choose meta implementation
        if self.options.meta_binary:
            meta = metasp.MetaspBinary(self)
        else:
            meta = metasp.MetaspPython(self)
        # get programs
        base, params, incremental = meta.get_incremental_program()
        # add to control
        self.control.add(METAUNSAT_BASE, [], base)
        self.control.add(METAUNSAT, params, incremental)

    #
    # exiting
    #

    def print_stats(self, interrupted=False, solved=True, copy_statistics=None):
        self.printer.print_stats(
            self.control, self.models, self.more_models, self.opt_models,
            self.options.non_optimal, self.options.stats,
            interrupted, solved, copy_statistics
        )

    def signal_on_solving(self):
        self.print_stats(interrupted=True)
        self.exit(1)

    def signal_on_not_solving(self):
        self.print_stats(
            interrupted=True, copy_statistics=self.control_proxy.statistics
        )
        self.exit(1)

    def signal_after_solving(self):
        self.print_stats()

    def exit(self, code):
        self.exited = True
        sys.exit(code)

    def end(self):
        self.print_stats()
        raise EndException


    #
    # run()
    #

    def run(self):

        # controllers
        general = controller.GeneralController(self)
        optimal = controller.GeneralControllerHandleOptimal(self)
        enumeration = controller.EnumerationController(self)
        self.on_optimal = on_optimal = controller.OnOptimalController(self)
        # MethodController
        if self.options.solving_mode == "weak":
            method = controller.ApproxMethodController(self)
            self.on_optimal = on_optimal
        elif self.options.solving_mode == "heuristic":
            method = controller.HeurMethodController(self)
        elif self.options.meta in [META_SIMPLE]:
            method = controller.MetaMethodController(self)
        else:
            if self.options.ground_once:
                method = controller.GroundOnceMethodController(self)
            else:
                method = controller.GroundManyMethodController(self)
        if self.options.improve_limit is not None:
            method = controller.ImproveLimitController(self, method)

        # loop
        try:
            # START
            general.start()
            optimal.start()
            method.start() # Approx and Meta finish here
            self.printer.do_print("Solving...")
            while True:
                # START_LOOP
                method.start_loop()
                # SOLVE
                method.solve()
                if self.solving_result == SATISFIABLE:
                    # SAT
                    general.sat()
                    optimal.sat()
                elif self.solving_result == UNSATISFIABLE:
                    # UNSAT
                    general.unsat()
                    method.unsat()
                    enumeration.unsat()
                    optimal.unsat()
                    on_optimal.unsat()
                elif self.solving_result == UNKNOWN:
                    # UNKNOWN
                    general.unknown()
                    method.unsat()
                    optimal.unknown()
                    on_optimal.unsat()
                # END_LOOP
                general.end_loop()
        except RuntimeError as e:
            if not self.exited:
                self.printer.print_error("ERROR (clingo): {}".format(e))
            sys.exit(1)
        except EndException as e:
            # END
            pass





from __future__ import print_function
import clingo
import threading
import sys
import signal
import copy
from . import clingo_stats

# defines
INTERRUPT  = """*** Info : ({}): INTERRUPTED by signal!
UNKNOWN

INTERRUPTED  : 1"""

SUMMARY_STR = """Calls        : 1
Time         : 0.000s (Solving: 0.00s 1st Model: 0.00s Unsat: 0.00s)
CPU Time     : 0.000s"""

STATS_STR = """
Choices      : 0
Conflicts    : 0        (Analyzed: 0)
Restarts     : 0
Problems     : 0        (Average Length: 0.00 Splits: 0)
Lemmas       : 0        (Deleted: 0)
  Binary     : 0        (Ratio:   0.00%)
  Ternary    : 0        (Ratio:   0.00%)
  Conflict   : 0        (Average Length:    0.0 Ratio:   0.00%)
  Loop       : 0        (Average Length:    0.0 Ratio:   0.00%)
  Other      : 0        (Average Length:    0.0 Ratio:   0.00%)
Backjumps    : 0        (Average:  0.00 Max:   0 Sum:      0)
  Executed   : 0        (Average:  0.00 Max:   0 Sum:      0 Ratio:   0.00%)
  Bounded    : 0        (Average:  0.00 Max:   0 Sum:      0 Ratio: 100.00%)

Variables    : 0        (Eliminated:    0 Frozen:    0)
Constraints  : 0        (Binary:   0.0% Ternary:   0.0% Other:   0.0%)i
"""

class ClingoSignalHandler:

    def __init__(self, control,
                 name="",
                 print_after_solving=False,
                 function_after_solving=None,
                 function_on_solving=None,
                 function_on_not_solving=None,
                 function_on_not_solved=None
                ):
        # public
        self.statistics = None
        self.function_after_solving = function_after_solving
        self.function_on_solving = function_on_solving
        self.function_on_not_solving = function_on_not_solving
        self.function_on_not_solved = function_on_not_solved
        # private
        self.interrupted = False
        self.control = control
        self.solved = False
        self.name = name
        self.print_after_solving = print_after_solving
        if self.function_after_solving is None:
            self.function_after_solving = self.after_solving
        if self.function_on_solving is None:
            self.function_on_solving = self.on_solving
        if self.function_on_not_solving is None:
            self.function_on_not_solving = self.on_not_solving
        if self.function_on_not_solved is None:
            self.function_on_not_solved = self.on_not_solved
        self.condition = threading.Condition()
        self.solving = False
        self.result = None
        # signal handling
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGINT, self.signal_handler)

    #
    # private: printing funtions
    #

    def do_print_stats(self, statistics):
        print(clingo_stats.Stats().summary(statistics))
        print(clingo_stats.Stats().statistics(statistics))

    def after_solving(self):
        self.do_print_stats(self.control.statistics)

    def on_solving(self):
        print(INTERRUPT.format(self.name))
        self.do_print_stats(self.control.statistics)
        sys.exit(1)

    def on_not_solving(self):
        print(INTERRUPT.format(self.name))
        statistics = self.control.statistics
        if self.statistics is not None:
            statistics = self.statistics
        self.do_print_stats(statistics)
        sys.exit(1)

    def on_not_solved(self):
        print(INTERRUPT.format(self.name))
        print(SUMMARY_STR)
        print(STATS_STR)
        sys.exit(1)

    #
    # signal handling
    #

    # private
    def signal_handler(self, signum, frame):
        if self.solving:
            self.control.interrupt()
            self.interrupted = True
        elif self.solved:
            self.function_on_not_solving()
        else:
            self.function_on_not_solved()

    # private
    def stop(self, result):
        self.result = result
        # notify
        with self.condition:
            self.condition.notify()

    # private
    def do_solve(self, control, *args, **kwargs):
        with self.condition:
            with control.solve(
                async=True, on_finish=self.stop, *args, **kwargs
            ) as handle:
                # In Python 2, Condition.wait() isn't interruptible when called without a timeout.
                # In Python 3, infinite timeouts lead to overflow errors.
                # To accomodate for both Python versions, a switch on the timeout is necessary.
                # More information available at: https://bugs.python.org/issue8844
                self.condition.wait(timeout = float("inf") if sys.version_info[0] < 3 else None)
                handle.wait()

    # public
    def solve(self, *args, **kwargs):
        self.solving = True
        # self.control.solve(*args, **kwargs)
        self.do_solve(self.control, *args, **kwargs)
        self.solved = True
        self.solving = False
        if self.interrupted:
            self.function_on_solving()
        elif self.print_after_solving:
            self.function_after_solving()
        return self.result

    # public
    # asyncronous solve that can be interrupted (control object is a parameter)
    def single_solve(self, control, *args, **kwargs):
        self.do_solve(control, *args, **kwargs)
        return self.result

    # public
    def ground(self, *args):
        if self.solved:
            self.statistics = copy.deepcopy(
                self.control.statistics
            )
        self.control.ground(*args)
        self.statistics = None


class Test:

    def __init__(self, name=""):
        self.name = name # used only for printing the INTERRUPT string

    #
    # select where to stop
    #   stop = 0 normally
    #   stop = 1 on_solving     (solves for long)
    #   stop = 2 on_not_solved  (infinite loop)
    #   stop = 3 on_not_solving (grounds for long, signals do not work)
    #   stop = 4 on_not_solving (infinite loop)
    #
    def run(self, stop=0):

        # start
        options = ["--stats"]
        control = clingo.Control(options)
        clingo_proxy = ClingoSignalHandler(
            control,
            self.name,
            print_after_solving=True
            #print_after_solving=False
        )

        # base program
        const = "#const n=8."
        if stop == 1:
            const = "#const n=10."
        control.add("base", [], const)
        control.add("base", [], """
pigeon(1..n+1). box(1..n).
1 { in(X,Y) : box(Y) } 1 :- pigeon(X).
:- 2 { in(X,Y) : pigeon(X) },  box(Y), not x.
#external x.
""")

        # infinite loop?
        while True:
            if stop == 2:
                pass
            else:
                break

        # ground and solve
        clingo_proxy.ground([("base", [])])
        print(clingo_proxy.solve(
            on_model=lambda x: print(x.symbols(shown=True))
        ))

        if stop == 3:
            control.add("big", [], "#const m=300. a(1..m,1..m,1..m).")
            clingo_proxy.ground([("big", [])])

        # infinite loop?
        while True:
            if stop == 4:
                pass
            else:
                break

if __name__ == "__main__":
    stop = 0
    #stop = 1
    #stop = 2
    #stop = 3
    #stop = 4
    Test("test").run(stop)




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

#!/usr/bin/python

import os
import sys
from .ply import yacc
from .ply.lex import LexToken
from .spec_lexer import Lexer
from ..utils import printer
from ..utils import utils
from . import ast


#
# defines
#

# programs
EMPTY     = utils.EMPTY
BASE      = utils.BASE
SPEC      = utils.SPEC
GENERATE  = utils.GENERATE
PREFP     = utils.PREFP
HEURISTIC = utils.HEURISTIC
APPROX    = utils.APPROX
UNSATP    = utils.UNSATP

# predicate names
DOM        = utils.DOM
GEN_DOM    = utils.GEN_DOM

# translation tokens
HASH_SEM   = utils.HASH_SEM

# errors
ERROR_PREFIX = "syntax error, "
ERROR_PREFERENCE = ERROR_PREFIX + "preference statement in non base program\n"
ERROR_OPTIMIZE  = ERROR_PREFIX + "optimize statement in non base program\n"
ERROR_PREFERENCE_NAME = ERROR_PREFIX + "incorrect preference name\n"
ERROR_SYNTAX = ERROR_PREFIX + "unexpected {}\n"
ERROR_CLINGO_STATEMENT = ERROR_PREFIX + """\
clingo optimize statement mixed with a preference specification\n"""

# more
PROGRAM      = "PROGRAM"
CODE         = "CODE"
PREFERENCE   = "PREFERENCE"
OPTIMIZE     = "OPTIMIZE"
STDIN        = "-"
END          = "end."
CLINGOPATH   = "CLINGOPATH"
ASPRIN_LIB   = "asprin_lib.lp"
ASPRIN_LIB_RELATIVE = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                   "..", "..", ASPRIN_LIB)
# WARNING: ASPRIN_LIB_RELATIVE must be changed if 
#          asprin_lib.lp location relative to this file changes

# clingo minimize statements
MINIMIZE_NAME = "clingo"
MINIMIZE_TYPE = "clingo_minimize"

#
# Program
#
class Program(object):

    def __init__(self, string):
        self.__positions = [] # list of utils.ProgramPositions
        self.__string = string 

    def get_string(self):
        return self.__string

    def get_positions(self):
        return self.__positions

    def extend_string(self, string):
        self.__string += "\n" + string

    def extend_positions(self, position):
        self.__positions.append(position)

#
#
# Ply Preference Specification Parser
#
#

class Parser(object):


    def __init__(self, underscores, options):

        # start lexer, parser, and printer
        self.options = options
        self.lexer  = Lexer(underscores, options)
        self.tokens = self.lexer.tokens
        #self.parser = yacc.yacc(module=self,debug=False)
        self.parser = yacc.yacc(module=self)
        self.printer = printer.Printer()

        # programs[name][type] is a Program
        self.p_statements, self.list = 0, []
        l = [BASE, GENERATE, SPEC, PREFP, HEURISTIC, APPROX, UNSATP]
        self.programs = dict([(i,dict()) for i in l])

        # base
        self.base      = ast.ProgramStatement()
        self.base.name = BASE
        self.base.type = EMPTY

        # others
        self.constants   = []
        self.included    = []
        self.error       = False
        self.position    = None
        self.element     = None
        self.filename    = None
        self.program     = None

        # clingo optimize statements
        self.preference_statement = False
        self.clingo_statement     = False


    #
    # AUXILIARY FUNCTIONS
    #

    def __handle_clingo_statements(self, preference, p, init, end):
        if preference:
            self.preference_statement = True
        else:
            self.clingo_statement = True
        if self.preference_statement and self.clingo_statement:
            self.__syntax_error(p, init, end, ERROR_CLINGO_STATEMENT)


    def __get_col(self, data, pos):
        return pos - data.rfind('\n', 0, pos)


    # get Location from non terminal p between init and end
    # p[init] and p[end] should be either terminals, or the error token 
    def __get_location(self, p, init, end):
        filename   = self.filename
        line       = p.lineno(init)
        col_ini    = self.__get_col(self.lexer.lexer.lexdata, p.lexpos(init))
        line_extra = p.lineno(end)
        col_end    = self.__get_col(self.lexer.lexer.lexdata, p.lexpos(end))
        if isinstance(p[end], str):
            col_end += len(p[end])
        elif isinstance(p[end], LexToken):
            col_end += len(p[end].value)
        elif init == end:
            col_end += 1
        return utils.Location(filename, line, col_ini, line_extra, col_end)


    # p[init] and p[end] should be either terminals, or the error token 
    def __syntax_error(self, p, init, end, string):
        self.error = True
        location = self.__get_location(p, init, end)
        self.printer.print_error_location(location, string)


    def __update_program(self, program, type, string, position=None):
        # consider only the program names in self.programs
        dictionary = self.programs.get(program)
        if dictionary is None:
            return
        # update the program type
        program = dictionary.get(type)
        if program is None:
            program = Program(string)
            dictionary[type] = program
        else:
            program.extend_string(string)
        if position is not None:
            program.extend_positions(position)


    def __generate_programs(self):

        underscores = "_" + ("_" * self.lexer.get_underscores())
        ast.Statement.underscores = underscores
        program, type = BASE, EMPTY

        # add elements of the list
        for i in self.list:
            if i[0] == CODE:
                code = i[1]
                # if base: add END to mark the end of the code
                if program == BASE and type == EMPTY:
                    code += underscores + END
                self.__update_program(program, type, code, i[2])
            if i[0] == PREFERENCE or i[0] == OPTIMIZE:
                # translate statement
                self.__update_program(SPEC, EMPTY, i[1].str())
            if i[0] == PROGRAM:
                program, type = i[1].name, i[1].type

        # adding generation of domains
        self.__update_program(GENERATE,EMPTY,"\n".join(ast.Statement.domains))

        # adding specification
        out = ""
        if ast.PStatement.bfs:
            out +=  ast.BF_ENCODING.replace("##",underscores)
        out += "\n" + ast.TRUE_ATOM.replace("##",underscores)
        out += "\n" + ast.PREF_DOM_RULE.replace("##",underscores)
        self.__update_program(SPEC,EMPTY,out)

        return self.programs, underscores


    def __parse_file(self, filename):
        # set variables
        self.filename = filename
        self.program  = BASE
        self.element  = ast.Element()
        self.position = utils.ProgramPosition(self.filename, 1, 1)
        # add #program base to list
        self.list.append((PROGRAM, self.base))
        # prepare lexer
        self.lexer.new_file(filename)
        # handle file descriptor, and parse
        fd = sys.stdin if filename == STDIN else open(filename)
        self.parser.parse(fd.read(), self.lexer.lexer) # parses into self.list
        fd.close()


    def __search_in_clingopath(self, file):
        path = os.environ.get(CLINGOPATH)
        if path is None:
            return None
        full = os.path.join(os.path.dirname(path), file)
        if not os.path.isfile(full):
            return None
        return file


    def __parse_included_files(self, files):
        while True:
            included, self.included = self.included, []
            for i in included: # (filename, fileorigin)
                file = i[0]
                if not os.path.isfile(file): # look in the directory
                    file = os.path.join(os.path.dirname(i[1]), i[0])
                    if not os.path.isfile(file): # look in CLINGOPATH
                        file = self.__search_in_clingopath(i[0])
                        if file is None:
                            self.printer.error_included_file(i[0], i[2])
                            self.error = True
                            continue
                abs_file = os.path.abspath(file)
                if abs_file in [j[1] for j in files]: 
                    self.printer.warning_included_file(file, i[2])
                else:
                    files.append((file, abs_file))
                    self.__parse_file(file)
            if self.included == []:
                return


    #
    # Input:  options [list of files, bool for including asprin lib]
    # Output: translation, underscores, constants, and programs with shown 
    #
    def parse_files(self):

        # input files
        files = self.options['files']
        for i in files:
            if i[0]=="-":
                self.__parse_file(STDIN)
            else:
                self.__parse_file(i[0])

        # included files
        self.__parse_included_files(files)

        # asprin_lib.lp
        #filenames = [os.path.basename(i[0]) for i in files]
        #if self.options['asprin-lib'] and ASPRIN_LIB not in filenames:
        if self.options['asprin-lib']:
            if os.path.isfile(ASPRIN_LIB):
                file = ASPRIN_LIB
            else:
                file = ASPRIN_LIB_RELATIVE
            self.__parse_file(file)

        # errors
        if self.lexer.get_error() or self.error:
            raise Exception("parsing failed")

        # return
        programs, underscores = self.__generate_programs()
        return programs, underscores, self.constants, self.lexer.get_show()



    #
    # Syntax:
    #   A preference statement has form:
    #     #preference(name,type) { E1; ...; En } [ : B ].
    #   where
    #     name and type are terms,
    #     Ei are preference elements, and
    #     B is a body of literals.
    #   A preference element has form:
    #     S1 [ >> ... >> Sn ] [ || S0 ] [ : B ]
    #   where
    #     Si is a set with weighted bodies of boolean formulas, 
    #     or with weighted naming atoms, or with a combination of both.
    #   A weighted body of boolean formulas has form:
    #     [T ::] BF1, ..., BFn
    #   or
    #     T ::
    #   where T is a tuple of terms, and
    #   BFi is a boolean formula of literals 
    #   (using 'not', '&', '|', '(' and ')').
    #   The second case is interpreted as:
    #     T :: #true
    #   A weighted naming atom has form:
    #     [T ::] **A
    #   where A is a term.
    #   An optimize statement has form:
    #     #optimize(name) [ : B ].
    #
    # Minor notes:
    #   (atom)             is not allowed (and never needed :)
    #   bitwise operator & is not allowed (Roland said that this operator
    #     will be eliminated from the clingo language)
    #
    # Comparison with minimize statements:
    #   asprin accepts boolean formulas
    #   asprin does not accept @ symbol
    #   asprin requires either 'T ::', or a body of boolean formulas, or 
    #     a naming atom, while clingo requires T
    #   if an element has no COLONS, asprin interprets it as a body of 
    #     boolean formulas, or a naming atom, while clingo interprets it as T
    # Translation: clingo elements
    #   T [: [ Body ] ]
    # without @ symbol, are translated into asprin elements of form
    #   T :: [ Body ]
    # The translation in the other direction is not always possible
    #

    #
    # START
    #

    precedence = (
        ('left', 'DOTS'),
        ('left', 'XOR'),
        ('left', 'QUESTION'),
        ('left', 'AND'),
        ('left', 'ADD', 'SUB'),
        ('left', 'MUL', 'SLASH', 'MOD'),
        ('right','POW'),
        ('right','POW_NO_WS'),
        ('left', 'UMINUS', 'UBNOT'),
        ('left', 'BFVBAR'),
        ('left', 'BFAND'),
        ('left', 'BFNOT'),
    )

    start = 'program'    # the start symbol in our grammar

    #
    # PROGRAM
    #

    def p_program(self, p):
        """ program : program statement change_state CODE
                    | CODE
        """
        self.position.lines = self.lexer.lexer.lineno - self.position.line + 1
        if   len(p) == 5: 
            self.list.append((CODE,p[4],self.position))
        elif len(p) == 2: 
            self.list.append((CODE,p[1],self.position))

    def p_program_error(self, p):
        """ statement : error DOT
        """
        self.__syntax_error(p,1,1,ERROR_SYNTAX.format(p[1].value))

    def p_change_state(self, p):
        """ change_state :
        """
        p.lexer.pop_state()
        lexpos, lineno = self.lexer.lexer.lexpos, self.lexer.lexer.lineno
        self.lexer.set_code_start(lexpos)
        # position 
        col = lexpos - self.lexer.lexer.lexdata.rfind('\n', 0, lexpos)
        self.position = utils.ProgramPosition(self.filename, lineno, col)


    #
    # PREFERENCE STATEMENT
    #

    def p_statement_1(self, p):
        """ statement : PREFERENCE LPAREN term COMMA term RPAREN \
                        LBRACE elem_list RBRACE body DOT
                      | PREFERENCE LPAREN term COMMA term RPAREN \
                        LBRACE           RBRACE body DOT
        """
        # create preference statement
        s = ast.PStatement()
        self.p_statements += 1
        s.number = self.p_statements
        s.name     = p[3]
        s.type     = p[5]
        s.elements = p[8] if len(p) == 12 else []
        s.body     = p[len(p)-2]
        self.list.append((PREFERENCE,s)) # appends to self.list
        # restart element
        self.element = ast.Element()
        # error if not in base
        if self.program != BASE:
            self.__syntax_error(p,1,len(p)-1,ERROR_PREFERENCE)
        # error if there is also a clingo statement
        self.__handle_clingo_statements(True, p, 1, len(p)-1)


    #
    # BODY
    #

    def p_body(self, p):
        """ body : COLON litvec
                 |
        """
        p[0] = None
        if len(p)==3: p[0] = p[2]

    def p_litvec(self, p):
        """ litvec : litvec COMMA         ext_atom
                   | litvec COMMA     NOT ext_atom
                   | litvec COMMA NOT NOT ext_atom
                   |                      ext_atom
                   |                  NOT ext_atom
                   |              NOT NOT ext_atom
        """
        p[0] = p[1:]


    #
    # PREFERENCE ELEMENTS
    #

    def p_elem_list(self, p):
        """ elem_list : elem_list SEM elem body
                      |               elem body
        """
        # set self.element values & append to elem_list
        if len(p) == 5:
            self.element.sets = p[3][0]
            self.element.cond = p[3][1]
            self.element.body = p[4]
            p[0] = p[1] + [self.element]
        else:
            self.element.sets = p[1][0]
            self.element.cond = p[1][1]
            self.element.body = p[2]
            p[0] = [self.element]
        # restart element
        self.element = ast.Element()

    def p_elem(self, p):
        """ elem : elem_head
                 | elem_head COND weighted_body_set
        """
        if len(p) == 2:
            p[0] = (p[1],[])
        else:
            p[0] = (p[1],p[3])
        self.element.vars = self.element.all_vars

    def p_elem_head(self, p):
        """ elem_head : elem_head GTGT weighted_body_set
                      |                weighted_body_set
        """
        if len(p) == 4:
            p[0] = p[1] + [p[3]]
        else:
            p[0] = [p[1]]

    def p_weighted_body_set(self, p):
        """ weighted_body_set : LBRACE weighted_body_vec RBRACE
                              |        weighted_body
        """
        if len(p) == 4:
            p[0] = p[2]
        else:
            p[0] = [p[1]]

    def p_weighted_body_vec(Self,p):
        """ weighted_body_vec : weighted_body_vec SEM weighted_body
                              |                       weighted_body
        """
        if len(p) == 4:
            p[0] = p[1] + [p[3]]
        else:
            p[0] = [p[1]]


    #
    # WEIGHTED BODY
    #

    def p_weighted_body_1(self, p):
        """ weighted_body :                      bfvec_x
        """
        p[0] = ast.WBody(None,p[1])

    def p_weighted_body_3(self, p):
        """ weighted_body : ntermvec_x TWO_COLON bfvec_x
        """
        p[0] = ast.WBody(p[1],p[3])

    def p_weighted_body_5(self, p):
        """ weighted_body : ntermvec_x TWO_COLON
        """
        p[0] = ast.WBody(p[1],[["ext_atom",["true",["#true"]]]])

    def p_weighted_body_6(self, p):
        """ weighted_body :                      POW_NO_WS term
        """
        p[0] = ast.WBody(None,p[2],True)

    def p_weighted_body_7(self, p):
        """ weighted_body : ntermvec_x TWO_COLON POW_NO_WS term
        """
        p[0] = ast.WBody(p[1],p[4],True)


    #
    # VECTORS
    #

    def p_ntermvec_x(self, p):
        """ ntermvec_x : atomvec
                       | na_ntermvec
                       | atomvec COMMA na_ntermvec
        """
        p[0] = p[1:]

    def p_atomvec(self, p):
        """ atomvec : atom
                    | atomvec COMMA atom
        """
        p[0] = p[1:]
        if len(p)==4:
            self.atomvec.append(["ext_atom",["atom",p[3]]])
        else:
            self.atomvec = [["ext_atom",["atom",p[1]]]]

    def p_na_ntermvec(self, p):
        """ na_ntermvec : na_term
                        | na_term COMMA ntermvec
        """
        p[0] = p[1:]

    #
    #   """ bfvec_x : atomvec
    #               | na_bfvec
    #               | atomvec COMMA na_bfvec
    #   """
    #
    #   p[0] becomes a list
    #
    def p_bfvec_x_1(self, p):
        """ bfvec_x  : atomvec
        """
        p[0] = self.atomvec

    def p_bfvec_x_2(self, p):
        """ bfvec_x  : na_bfvec
        """
        p[0] = p[1]

    def p_bfvec_x_3(self, p):
        """ bfvec_x :  atomvec COMMA na_bfvec
        """
        p[0] = self.atomvec + p[3]

    def p_na_bfvec(self, p):
        """ na_bfvec : na_bformula COMMA bfvec
                     | na_bformula
        """
        if len(p)==4:
            p[0] = [p[1]] + p[3]
        else:
            p[0] = [p[1]]

    def p_bfvec(self, p):
        """ bfvec    : bfvec COMMA bformula
                     |             bformula
        """
        if len(p)==4:
            p[0] = p[1] + [p[3]]
        else:
            p[0] = [p[1]]


    #
    # BOOLEAN FORMULAS
    #

    #
    # using non reachable token NEVER for making the grammar LALR(1)
    # (atom) is not allowed
    #
    #    """ bformula :               ext_atom
    #                 |         paren_bformula
    #                 | bformula VBAR bformula %prec BFVBAR
    #                 | bformula AND  bformula %prec BFAND
    #                 |          NOT  bformula %prec BFNOT
    #
    #                 | LPAREN     identifier                      NEVER
    #                 | LPAREN     identifier LPAREN argvec RPAREN NEVER
    #                 | LPAREN SUB identifier                      NEVER
    #                 | LPAREN SUB identifier LPAREN argvec RPAREN NEVER
    #
    #        paren_bformula : LPAREN na_bformula RPAREN
    #
    #        na_bformula :            na_ext_atom
    #                    |         paren_bformula
    #                    | bformula VBAR bformula %prec BFVBAR
    #                    | bformula AND  bformula %prec BFAND
    #                    |          NOT  bformula %prec BFNOT
    #    """

    def p_bformula_1(self, p):
        """ bformula :               ext_atom
        """
        p[0] = p[1]

    def p_bformula_2(self, p):
        """ bformula :         paren_bformula
        """
        p[0] = p[1]

    def p_bformula_3(self, p):
        """ bformula : bformula VBAR bformula %prec BFVBAR
        """
        p[0] = ["or",[p[1],p[3]]]

    def p_bformula_4(self, p):
        """ bformula : bformula  AND bformula %prec BFAND
        """
        p[0] = ["and",[p[1],p[3]]]

    def p_bformula_5(self, p):
        """ bformula :           NOT bformula %prec BFNOT
        """
        p[0] = ["neg",[p[2]]]

    # unreachable
    def p_formula_6(self, p):
        """ bformula : LPAREN     identifier                      NEVER IF
                     | LPAREN     identifier LPAREN argvec RPAREN NEVER IF
                     | LPAREN SUB identifier                      NEVER IF
                     | LPAREN SUB identifier LPAREN argvec RPAREN NEVER IF
        """
        pass

    def p_paren_formula(self, p):
        """ paren_bformula : LPAREN na_bformula RPAREN
        """
        p[0] = p[2]

    def p_na_bformula_1(self, p):
        """ na_bformula :            na_ext_atom
        """
        p[0] = p[1]

    def p_na_bformula_2(self, p):
        """ na_bformula :         paren_bformula
        """
        p[0] = p[1]

    def p_na_bformula_3(self, p):
        """ na_bformula : bformula VBAR bformula %prec BFVBAR
        """
        p[0] = ["or",[p[1],p[3]]]

    def p_na_bformula_4(self, p):
        """ na_bformula : bformula  AND bformula %prec BFAND
        """
        p[0] = ["and",[p[1],p[3]]]

    def p_na_bformula_5(self, p):
        """ na_bformula :           NOT bformula %prec BFNOT
        """
        p[0] = ["neg",[p[2]]]


    #
    # NOT ATOM TERMS
    #

    def p_na_term(self, p):
        """ na_term : term      DOTS term
                    | term       XOR term
                    | term  QUESTION term
                    | term       ADD term
                    | term       SUB term
                    | term       MUL term
                    | term     SLASH term
                    | term       MOD term
                    | term       POW term
                    | term POW_NO_WS term
                    |            na_term_more
                    | many_minus na_term_more
                    | many_minus SUB identifier LPAREN argvec RPAREN %prec UMINUS
                    | many_minus SUB identifier
        """
        p[0] = p[1:]

    def p_na_term_more(self, p):
        """ na_term_more : BNOT term %prec UBNOT
                         | LPAREN tuplevec RPAREN
                         | AT identifier LPAREN   argvec RPAREN
                         | VBAR unaryargvec VBAR
                         | NUMBER
                         | STRING
                         | INFIMUM
                         | SUPREMUM
                         | variable
                         | ANONYMOUS
        """
        p[0] = p[1:]

    def p_many_minus(self, p):
        """many_minus : SUB
                      | many_minus SUB %prec UMINUS
        """
        p[0] = p[1:]


    #
    # (NOT ATOM) EXTENDED ATOMS
    #
    #   """ ext_atom : TRUE
    #                | FALSE
    #                | atom
    #                | term cmp term
    #   """
    #
    def p_ext_atom_1(self, p):
        """ ext_atom : TRUE
        """
        p[0] = ["ext_atom",["true",["#true"]]]

    def p_ext_atom_2(self, p):
        """ ext_atom : FALSE
        """
        p[0] = ["ext_atom",["false",["#false"]]]

    def p_ext_atom_3(self, p):
        """ ext_atom : atom
        """
        p[0] = ["ext_atom",["atom",p[1]]]

    def p_ext_atom_4(self, p):
        """ ext_atom : term cmp term
        """
        p[0] = ["ext_atom",["cmp",p[1:]]]


    #   """ na_ext_atom : TRUE
    #                   | FALSE
    #                   | term cmp term
    #   """
    def p_na_ext_atom_1(self, p):
        """ na_ext_atom : TRUE
        """
        p[0] = ["ext_atom",["true",["#true"]]]

    def p_na_ext_atom_2(self, p):
        """ na_ext_atom : FALSE
        """
        p[0] = ["ext_atom",["false",["#false"]]]

    def p_na_ext_atom_3(self, p):
        """ na_ext_atom : term cmp term
        """
        p[0] = ["ext_atom",["cmp",p[1:]]]


    #
    # VARIABLES
    #
    def p_variable(self, p):
        """ variable : VARIABLE
        """
        p[0] = p[1]
        self.element.all_vars.add(p[1])


    #
    # GRINGO expressions
    #

    def p_term(self, p):
        """ term : term      DOTS term
                 | term       XOR term
                 | term  QUESTION term
                 | term       ADD term
                 | term       SUB term
                 | term       MUL term
                 | term     SLASH term
                 | term       MOD term
                 | term       POW term
                 | term POW_NO_WS term
                 |            SUB term %prec UMINUS
                 |           BNOT term %prec UBNOT
                 |               LPAREN tuplevec RPAREN
                 |    identifier LPAREN   argvec RPAREN
                 | AT identifier LPAREN   argvec RPAREN
                 | VBAR unaryargvec VBAR
                 | identifier
                 | NUMBER
                 | STRING
                 | INFIMUM
                 | SUPREMUM
                 | variable
                 | ANONYMOUS
        """
        p[0] = p[1:]

    def __handle_sem(self, p):
        self.element.pooling = True
        if len(p)==4:
            if isinstance(p[1],list) and len(p[1])==2 and p[1][0]==HASH_SEM:
                  return [HASH_SEM,p[1][1] + [p[3]]]
            else: return [HASH_SEM,[p[1]]  + [p[3]]]
        return p[1]

    def p_unaryargvec(self, p):
        """ unaryargvec :                  term
                        |  unaryargvec SEM term
        """
        p[0] = self.__handle_sem(p)

    def p_ntermvec(self, p):
        """ ntermvec : term
                     | ntermvec COMMA term
        """
        p[0] = p[1:]

    def p_termvec(self, p):
        """ termvec : ntermvec
                    |
        """
        p[0] = p[1:]

    def p_tuple(self, p):
        """ tuple : ntermvec COMMA
                  | ntermvec
                  |          COMMA
                  |
        """
        p[0] = p[1:]

    def p_tuplevec(self, p):
        """ tuplevec :              tuple
                     | tuplevec SEM tuple
        """
        p[0] = self.__handle_sem(p)

    def p_argvec(self, p):
        """ argvec :            termvec
                   | argvec SEM termvec
        """
        p[0] = self.__handle_sem(p)

    def p_cmp(self, p):
        """ cmp :  GT
                |  LT
                | GEQ
                | LEQ
                |  EQ
                | NEQ
        """
        p[0] = p[1]

    def p_atom(self, p):
        """ atom :     identifier
                 |     identifier LPAREN argvec RPAREN
                 | SUB identifier
                 | SUB identifier LPAREN argvec RPAREN
        """
        p[0] = p[1:]


    def p_identifier(self, p):
        """ identifier : IDENTIFIER
        """
        p[0] = p[1]


    #
    # OPTIMIZE
    #

    def p_statement_2(self, p):
        """ statement : OPTIMIZE LPAREN term RPAREN body DOT
        """
        s = ast.OStatement()
        s.name     = p[3]
        s.body     = p[5]
        self.list.append((OPTIMIZE,s)) # appends to self.list
        if self.program != BASE: 
            self.__syntax_error(p,1,6,ERROR_OPTIMIZE)
        # error if there is also a clingo statement
        self.__handle_clingo_statements(True, p, 1, len(p)-1)


    #
    # PROGRAM
    #

    def p_statement_3(self, p):
        """ statement : PROGRAM identifier LPAREN ntermvec RPAREN DOT
                      | PROGRAM identifier DOT
        """
        s      = ast.ProgramStatement()
        s.name = ast.ast2str(p[2])
        s.type = ast.ast2str(p[4]) if len(p)==7 else EMPTY
        self.list.append((PROGRAM,s)) # appends to self.list
        self.program = s.name
        self.lexer.set_program((s.name,s.type))
        #TODO: Check if there are variables in ntermvec
        if len(p)==7 and p[2]==PREFP and len(p[4])!=1:
            self.__syntax_error(p,3,5,ERROR_PREFERENCE_NAME)


    #
    # CONST
    #

    def p_statement_4(self, p):
        """ statement : CONST identifier EQ term DOT
        """
        if self.program == BASE:
            self.constants.append((ast.ast2str(p[2]),ast.ast2str(p[4])))
        else:
            line = "#const {}={}.".format(ast.ast2str(p[2]), ast.ast2str(p[4]))
            location = self.__get_location(p,1,5)
            self.list.append((CODE, line, location.get_position()))


    #
    # INCLUDE
    #

    def p_statement_5(self, p):
        """ statement : INCLUDE STRING DOT
        """
        location = self.__get_location(p,1,len(p)-1)
        # (file name included, current file name, location)
        self.included.append((p[2][1:-1],self.filename,location))


    #
    # CLINGO OPTIMIZE STATEMENT (not used now)
    #

    def p_statement_6(self, p):
        """ statement : MINIMIZE LBRACE RBRACE DOT
                      | MAXIMIZE LBRACE RBRACE DOT
                      | MINIMIZE LBRACE min_elem_list RBRACE DOT 
                      | MAXIMIZE LBRACE max_elem_list RBRACE DOT 
        """
        # create preference statement
        s = ast.PStatement()
        self.p_statements += 1
        s.number = self.p_statements
        s.name     = MINIMIZE_NAME
        s.type     = MINIMIZE_TYPE
        s.elements = p[3] if len(p) == 6 else []
        s.body     = None
        self.list.append((PREFERENCE,s)) # appends to self.list
        # restart element
        self.element = ast.Element()
        # error if not in base
        if self.program != BASE:
            self.__syntax_error(p, 1, len(p)-1, ERROR_PREFERENCE)
        # add optimize statement
        s = ast.OStatement()
        s.name     = MINIMIZE_NAME
        s.body     = None
        self.list.append((OPTIMIZE,s)) # appends to self.list
        # error if there is also a preference specification
        self.__handle_clingo_statements(False, p, 1, len(p)-1)

    def p_min_elem_list(self, p):
        """ min_elem_list : min_elem_list SEM min_weighted_body
                          |                   min_weighted_body
        """
        # set self.element values & append to elem_list
        if len(p) == 4:
            self.element.sets = [[p[3]]]
            self.element.cond = []
            self.element.body = None
            p[0] = p[1] + [self.element]
        else:
            self.element.sets = [[p[1]]]
            self.element.cond = [] 
            self.element.body = None
            p[0] = [self.element]
        # restart element
        self.element = ast.Element()

    def p_min_weighted_body_1(self, p):
        """ min_weighted_body : min_weight min_tuple COLON bfvec_x
                              | min_weight min_tuple COLON
                              | min_weight min_tuple
        """
        if len(p) == 5:
            p[0] = ast.WBody(p[1]+p[2], p[4])
        else:
            p[0] = ast.WBody(p[1]+p[2], [["ext_atom",["true",["#true"]]]])

    def p_min_weight(self, p):
        """ min_weight : term AT term
                       | term
        """
        if len(p) == 4:
            p[0] = [p[1], ",", p[3]]
        else:
            p[0] = [p[1], ",", "1"] # by default, level 1

    def p_min_tuple(self, p):
        """ min_tuple : COMMA ntermvec
                      |
        """
        p[0] = []
        if len(p)==3:
            p[0] = p[1:]

    # copy from p_min_elem_list, replacing by max
    def p_max_elem_list(self, p):
        """ max_elem_list : max_elem_list SEM max_weighted_body
                          |                   max_weighted_body
        """
        # set self.element values & append to elem_list
        if len(p) == 4:
            self.element.sets = [[p[3]]]
            self.element.cond = []
            self.element.body = None
            p[0] = p[1] + [self.element]
        else:
            self.element.sets = [[p[1]]]
            self.element.cond = [] 
            self.element.body = None
            p[0] = [self.element]
        # restart element
        self.element = ast.Element()

    # copy from p_min_weighted_body, adding minus
    def p_max_weighted_body_1(self, p):
        """ max_weighted_body : min_weight min_tuple COLON bfvec_x
                              | min_weight min_tuple COLON
                              | min_weight min_tuple
        """
        if len(p) == 5:
            p[0] = ast.WBody(["-"]+p[1]+p[2], p[4])
        else:
            p[0] = ast.WBody(["-"]+p[1]+p[2], [["ext_atom",["true",["#true"]]]])


    #
    # ERROR
    #

    def p_error(self, p):
        return




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

#!/usr/bin/python
from __future__ import print_function
import os
import sys
import subprocess
from . import utils

PATH = os.path.dirname(os.path.realpath(__file__))
DIR = "--test-dir="
ALL = "--all"
OPTIONS = [
    [""],
    ["--delete-better"],
    ["--ground-once"],
    ["--release-last"],
    ["--no-opt-improving"],
    ["--volatile-improving"],
    ["--volatile-optimal"],
    ["--approximation=heuristic"],
    ["--approximation=heuristic --const-nb heuristic_aso=2"],
    ["--approximation=heuristic --const-nb heuristic_aso=2 --const-nb use_get_sequence=2 "],
    ["""--approximation=heuristic --const-nb heuristic_poset=2 --const-nb heuristic_aso=3 \
     --const-nb heuristic_pareto=2 --const-nb heuristic_and=2"""],
    ["--approximation=weak"],
    ["--approximation=weak --const-nb approx_aso=2"],
    ["--approximation=weak --const-nb approx_aso=2 --const-nb use_get_sequence=2 "],
    ["""--approximation=weak --const-nb approx_poset=2 --const-nb approx_aso=3 \
     --const-nb approx_pareto=2 --const-nb approx_and=2"""],
    ["--improve-limit=0,5"],
    ["--improve-limit=1,all,100"],
    ["--on-opt-heur=+,p,-1,sign --on-opt-heur=-,p,1,sign"],
    ["--on-opt-heur=+,s,1,true --on-opt-heur=-,s,1,false"],
    ["--meta=simple"],
    ["--meta=simple,bin"],
    ["--meta=combine"],
    ["--meta=combine,bin"],
]

EXCLUDE = {}

EXCLUDE["--on-opt-heur=+,p,-1,sign --on-opt-heur=-,p,1,sign"] = [
    os.path.join(PATH, "program_parser", "basic", "test001.lp"), # uses --approximation=heuristic
    os.path.join(PATH, "program_parser", "basic", "test002.lp"), # uses --approximation=heuristic
]

EXCLUDE["--meta=simple"] = [
    os.path.join(PATH, "asprin_lib", "test022.lp"), # too hard
    os.path.join(PATH, "asprin_lib", "test023.lp"), # too hard
    os.path.join(PATH, "asprin_lib", "test024.lp"), # too hard
    os.path.join(PATH, "asprin_lib", "test025.lp"), # too hard
    os.path.join(PATH, "solver", "solver", "test008.lp"), # too hard
    os.path.join(PATH, "solver", "solver", "test009.lp"), # too hard
    os.path.join(PATH, "program_parser", "visitor", "test002.lp"), # no error with --meta
]
EXCLUDE[ "--meta=simple,bin"] = EXCLUDE["--meta=simple"]
EXCLUDE[    "--meta=combine"] = EXCLUDE["--meta=simple"]
EXCLUDE["--meta=combine,bin"] = EXCLUDE["--meta=simple"]

EXCLUDE["--preference-unsat $asprin/mine/asprin_lib_unsat.lp"] = [
    os.path.join(PATH, "solver", "solver", "test002.lp"),           # adds new preference programs
    os.path.join(PATH, "program_parser", "visitor", "test002.lp"),  # adds new preference programs
    os.path.join(PATH, "spec_parser", "spec_lexer", "test005.lp"),  # adds new preference programs
    os.path.join(PATH, "spec_parser", "spec_parser", "test026.lp"), # adds new preference programs
]

add_option = False
# to add one option to all OPTIONS, uncomment the next line and set option below
#add_option = True
option = "--preference-unsat $asprin/mine/asprin_lib_unsat.lp"
#
option_exclude = EXCLUDE[option]
if add_option:
    for i in OPTIONS:
        if i == option:
            continue
        if i[0] in EXCLUDE:
            EXCLUDE[i[0] + " " + option] = option_exclude + EXCLUDE[i[0]]
        else:
            EXCLUDE[i[0] + " " + option] = option_exclude
        i[0] += " " + option

# uncomment to only test --preference-unsat
#OPTIONS = [["--preference-unsat $asprin/mine/asprin_lib_unsat.lp"]]

class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)

class Tester:

    def exclude(self, options, _file):
        try:
            if _file in EXCLUDE[" ".join(options)]:
                return True
        except:
            pass
        return False

    def run(self, dir, options):
        errors, error = False, False
        for i in sorted(os.listdir(dir)):
            abs_i = os.path.join(dir,i)
            if self.exclude(options, abs_i):
                continue
            if os.path.isdir(abs_i):
                error = self.run(os.path.join(dir, i), options)
            elif str(abs_i)[-3:] == ".lp":
                with open(abs_i, 'r') as f:
                    test = utils.Test(f.read(), options)
                print("Testing {}...".format(abs_i))
                import tempfile
                tmp = tempfile.TemporaryFile()
                with cd(dir):
                    subprocess.call(test.command, stdout=tmp,
                                    stderr=subprocess.STDOUT, shell=True)
                    tmp.seek(0)
                    output = tmp.read()
                    if isinstance(output, bytes):
                        output = output.decode()
                    result = utils.Result(output)
                error = result.compare(test)
            if error:
                errors = True
        return errors

def main(args):
    path = PATH
    for i in args:
        if i.startswith(DIR):
            path = i[len(DIR):]
            args.remove(i)
            break
    if ALL in args:
        options = OPTIONS
    else:
        options = [args]
    global_errors = []
    for opt in options:
        print("Options = {}".format(opt))
        errors = Tester().run(path, opt)
        if errors:
            print("ERROR: There were errors in the tests")
            global_errors.append(str(opt))
        else:
            print("OK: All tests were successful")
    if len(options) <= 1:
        return
    if global_errors:
        e = " ".join(global_errors)
        print("\nSUMMARY: There were errors with options: {}".format(e))
    else:
        print("\nSUMMARY: There were no errors.")

if __name__ == "__main__":
    main()




from setuptools import setup, find_packages
import sys, os, codecs, re

HERE = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()

META_FILE = read(os.path.join("asprin", "__init__.py"))

def find_meta(meta):
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))

setup(
    name = find_meta("package"),
    version = find_meta("version"),
    description = find_meta("description"),
    author = find_meta("author"),
    author_email = find_meta("email"),
    url = find_meta("url"),
    license = find_meta("license"),
    long_description = read("README.md"),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console', 
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    keywords='logic answer set programming preference optimization',
    packages=find_packages('.'),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'asprin=asprin.asprin:main',
        ],
    },
)
